/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>
#include "kiba.h"
#include "kiba-win.h"
#include "kiba-debug.h"
#include "kiba-key-file.h"
#include "kiba-object.h"
#include "kiba-inline-pixbufs.h"
#include "kiba-desktop-icon.h"

typedef enum {
	KIBA_DESKTOP_TYPE_ENTRY = 0,
	KIBA_DESKTOP_NAME_ENTRY,
	KIBA_DESKTOP_COMMENT_ENTRY,
	KIBA_DESKTOP_EXEC_ENTRY,
	KIBA_DESKTOP_URL_ENTRY,
	KIBA_DESKTOP_PATH_ENTRY,
	KIBA_DESKTOP_ICON_ENTRY,
	KIBA_DESKTOP_GROUP_ENTRY,
	KIBA_DESKTOP_CLASS_ENTRY
}KibaDesktopKey;

enum {
	FILE_CHANGED_EVENT = 0,
	SIG_DESKTOP_ICON_LAST
};
static guint desktop_icon_signals[SIG_DESKTOP_ICON_LAST] = { 0 };

G_DEFINE_TYPE (KibaDesktopIcon, kiba_desktop_icon, G_TYPE_OBJECT)

static KibaDesktopIconType
kiba_desktop_icon_get_type_from_string (const char *string)
{
	static const char *icon_types [] = {
		"Application", "Link", "Directory", "Reminder"
	};

	gint i;

	g_return_val_if_fail (string != NULL, KIBA_DESKTOP_TYPE_INVALID);

	for (i = 0; i < G_N_ELEMENTS(icon_types); i++) {
		if (strcmp (string, icon_types[i]) == 0) {
			return i;
		}
	}

	return KIBA_DESKTOP_TYPE_INVALID;
}

static gboolean
kiba_desktop_icon_load_file (KibaDesktopIcon *icon,
		const gchar *uri,
		GError **error)
{
	static const gchar file_uri_prefix[] = "file://";
	static const gchar desktop_entry_group[] = "Desktop Entry";
	static const gchar desktop_file_suffix[] = ".desktop";
	static const struct { const gchar *string; KibaDesktopKey key; size_t offset; } desktop_data[] = {
		{ G_KEY_FILE_DESKTOP_KEY_TYPE, KIBA_DESKTOP_TYPE_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.type) },
		{ G_KEY_FILE_DESKTOP_KEY_NAME, KIBA_DESKTOP_NAME_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.name) },
		{ G_KEY_FILE_DESKTOP_KEY_COMMENT, KIBA_DESKTOP_COMMENT_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.comment) },
		{ G_KEY_FILE_DESKTOP_KEY_EXEC, KIBA_DESKTOP_EXEC_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.exec) },
		{ G_KEY_FILE_DESKTOP_KEY_URL, KIBA_DESKTOP_URL_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.exec) },
		{ G_KEY_FILE_DESKTOP_KEY_PATH, KIBA_DESKTOP_PATH_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.exec) },
		{ G_KEY_FILE_DESKTOP_KEY_ICON, KIBA_DESKTOP_ICON_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.icon) },
		{ KIBA_KEY_FILE_DESKTOP_KEY_GROUP, KIBA_DESKTOP_GROUP_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.group) }
	};
	static const gchar *desktop_keys[G_N_ELEMENTS (desktop_data)];

	GKeyFile *key_file;
	gchar *filename = NULL;
	gchar *s;
	gint i;

	for (i = 0; i < G_N_ELEMENTS (desktop_keys); i++) {
		desktop_keys[i] = desktop_data[i].string;
	}

	if (!g_file_test(uri, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_IS_SYMLINK)) {
		DEBUG_ERROR ("'%s' is not a valid File\n", uri);
		return FALSE;
	}
	if (!g_str_has_suffix(uri, desktop_file_suffix)) {
		DEBUG_ERROR ("'%s' is not a .desktop File\n", uri);
		return FALSE;
	}

	if (g_file_test(uri, G_FILE_TEST_IS_SYMLINK)) {
		filename = g_file_read_link (uri, error);
		if (filename == NULL)
			return FALSE;
	}
	else if (g_str_has_prefix (uri, file_uri_prefix)) {
		filename = g_strdup(uri + strlen (file_uri_prefix));
	}
	else {
		filename = g_strdup(uri);
	}

	if (!kiba_key_file_add_missing_desktop_entries (filename,
					desktop_entry_group,
					desktop_keys,
					G_N_ELEMENTS (desktop_keys))) {
		g_free (filename);
		return FALSE;
	}

	key_file = g_key_file_new ();
	if (!g_key_file_load_from_file (key_file,
			filename, G_KEY_FILE_KEEP_COMMENTS, error)) {
		g_free (filename);
		return FALSE;
	}

	for (i = 0; i < G_N_ELEMENTS (desktop_data); i++) {
		s = g_key_file_get_string (key_file, desktop_entry_group, desktop_data[i].string, error);
		g_clear_error (error);
		*error = NULL;

		if (!s) {
			continue;
		}
		
		switch (desktop_data[i].key) {
			case KIBA_DESKTOP_TYPE_ENTRY:
				G_STRUCT_MEMBER (gint , icon, desktop_data[i].offset) =
					kiba_desktop_icon_get_type_from_string (s);
				break;
			case KIBA_DESKTOP_NAME_ENTRY:
			case KIBA_DESKTOP_COMMENT_ENTRY: {
				gchar *translated;

				translated = g_key_file_get_locale_string (key_file,
							desktop_entry_group,
							desktop_data[i].string,
							NULL,
							error);
				g_clear_error (error);
				*error = NULL;

				if (translated) {
					g_free (s);
					G_STRUCT_MEMBER (gchar *, icon, desktop_data[i].offset) = translated;
				}
				else
					G_STRUCT_MEMBER (gchar *, icon, desktop_data[i].offset) = s;
				break;
			}
			case KIBA_DESKTOP_EXEC_ENTRY:
			case KIBA_DESKTOP_ICON_ENTRY:
			case KIBA_DESKTOP_GROUP_ENTRY:
				G_STRUCT_MEMBER (gchar *, icon, desktop_data[i].offset) = s;
				break;
			case KIBA_DESKTOP_URL_ENTRY:
			case KIBA_DESKTOP_PATH_ENTRY: {
				gchar *command;
				
				if (G_STRUCT_MEMBER (gchar *, icon, desktop_data[i].offset)) {
					break;
				}
				
				// TODO: custom filemanager
				// TODO: custom browser
				command = g_strdup_printf ("nautilus --no-desktop %s", s);
				g_free (s);
				G_STRUCT_MEMBER (gchar *, icon, desktop_data[i].offset) = command;
				break;
			}
		}
	}
	g_key_file_free (key_file);

	if (icon->entry.uri) {
		g_free (icon->entry.uri);
	}
	icon->entry.uri = filename;

	return TRUE;
}

/// public methods

/**
 * kiba_desktop_icon_set_file:
 * @desktop_icon: the #KibaDesktopIcon instance.
 * @desktop_file: path to a .desktop file.
 * 
 * Returns: TRUE on success, otherwhise FALSE is returned.
 **/
gboolean
kiba_desktop_icon_set_file (KibaDesktopIcon *desktop_icon,
		const gchar *desktop_file)
{
	g_return_if_fail (KIBA_IS_DESKTOP_ICON(desktop_icon));
	g_return_if_fail (desktop_file != NULL);
	
	GError *error = NULL;
	
	if (desktop_icon->entry.name) {
		g_free (desktop_icon->entry.name);
		desktop_icon->entry.name = NULL;
	}
	if (desktop_icon->entry.comment) {
		g_free (desktop_icon->entry.comment);
		desktop_icon->entry.comment = NULL;
	}
	if (desktop_icon->entry.exec) {
		g_free (desktop_icon->entry.exec);
		desktop_icon->entry.exec = NULL;
	}
	if (desktop_icon->entry.icon) {
		g_free (desktop_icon->entry.icon);
		desktop_icon->entry.icon = NULL;
	}
	if (desktop_icon->entry.group) {
		g_free (desktop_icon->entry.group);
		desktop_icon->entry.group = NULL;
	}
	
	if (!kiba_desktop_icon_load_file(desktop_icon, desktop_file, &error)) {
		DEBUG_ERROR ("Failed to parse desktop File '%s'\n", desktop_file);
		if (error != NULL) {
			DEBUG_ERROR ("%s\n", error->message);
			g_free (error);
		}
		return FALSE;
	}
	
	if (!desktop_icon->entry.group) {
		desktop_icon->entry.group = g_strdup("Default");
	}
	
	g_signal_emit (desktop_icon,
		desktop_icon_signals[FILE_CHANGED_EVENT], 0, desktop_file);

	return TRUE;
}

/**
 * kiba_desktop_icon_get_file:
 * @desktop_icon: the #KibaDesktopIcon instance.
 * Returns: path to a .desktop file.
 **/
const gchar*
kiba_desktop_icon_get_file (KibaDesktopIcon *desktop_icon)
{
	g_return_val_if_fail (KIBA_IS_DESKTOP_ICON(desktop_icon), NULL);
	return desktop_icon->entry.uri;
}

/**
 * kiba_desktop_icon_write_file:
 * @desktop_icon: #KibaDesktopIcon instance.
 * 
 * Write #KibaDesktopIcon instance back to the
 * previously set desktop file path.
 */
void
kiba_desktop_icon_write_file (KibaDesktopIcon *desktop_icon)
{
}

/// GObject implementation

static void
kiba_desktop_icon_dispose (GObject *g_object)
{
	KibaDesktopIcon *desktop_icon = KIBA_DESKTOP_ICON(g_object);
	GList *l;

	if (desktop_icon->entry.uri) {
		g_free (desktop_icon->entry.uri);
		desktop_icon->entry.uri = NULL;
	}
	if (desktop_icon->entry.name) {
		g_free (desktop_icon->entry.name);
		desktop_icon->entry.name = NULL;
	}
	if (desktop_icon->entry.comment) {
		g_free (desktop_icon->entry.comment);
		desktop_icon->entry.comment = NULL;
	}
	if (desktop_icon->entry.exec) {
		g_free (desktop_icon->entry.exec);
		desktop_icon->entry.exec = NULL;
	}
	if (desktop_icon->entry.icon) {
		g_free (desktop_icon->entry.icon);
		desktop_icon->entry.icon = NULL;
	}
	if (desktop_icon->entry.group) {
		g_free (desktop_icon->entry.group);
		desktop_icon->entry.group = NULL;
	}
}

/// Class Functions

static void
kiba_desktop_icon_class_init (KibaDesktopIconClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	
	gobject_class->dispose = kiba_desktop_icon_dispose;
	
	/**
	 * KibaDesktopIcon::file-changed:
	 * @desktop_icon: the object on which the signal is emitted
	 *
	 * The file-changed signal is emitted when a new file has been set
	 * on @desktop_icon.
	 */
	desktop_icon_signals[FILE_CHANGED_EVENT] =
		g_signal_new ("file-changed",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaDesktopIconClass, file_changed),
				NULL, NULL,
				g_cclosure_marshal_VOID__STRING,
				G_TYPE_NONE, 1,
				G_TYPE_CHAR);
}

static void
kiba_desktop_icon_init (KibaDesktopIcon *desktop_icon)
{
	desktop_icon->entry.type = KIBA_DESKTOP_TYPE_INVALID;
	desktop_icon->entry.uri = NULL;
	desktop_icon->entry.name = NULL;
	desktop_icon->entry.comment = NULL;
	desktop_icon->entry.exec = NULL;
	desktop_icon->entry.icon = NULL;
	desktop_icon->entry.group = NULL;
}

KibaDesktopIcon*
kiba_desktop_icon_new (void)
{
	KibaDesktopIcon *icon;

	icon = g_object_new (KIBA_TYPE_DESKTOP_ICON, NULL);

	return icon;
}
