/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-utils
 * @short_description: General utility methods.
 * @see_also:
 *
 * General utility methods.
 */

#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <gtk/gtk.h>
#include <libintl.h>
#include <sys/stat.h>

#include "kiba-utils.h"
#include "kiba-macros.h"
#include "kiba-dialog.h"
#include "kiba-debug.h"

#define PROC_PATH     "/proc"
#define CPU_INFO_PATH PROC_PATH"/cpuinfo"
#define COMMAND_FILE  "/cmdline"

#define _(string)  gettext (string)

gboolean
kiba_utils_recursive_mkdir (const gchar *path)
{
	// I cannot believe this isn't in some library somewhere

	gchar **split_path;
	gchar *so_far;
	gint i;

	split_path = g_strsplit (path, G_DIR_SEPARATOR_S, 0);

	so_far = g_strdup ("/");
	for (i = 0; split_path[i] != NULL; i++) {
		gchar *to_check;

		to_check = g_build_filename (so_far, split_path[i], NULL);
		g_free (so_far);

		if (!g_file_test (to_check, G_FILE_TEST_EXISTS)) {
			if (mkdir (to_check, S_IRWXU) != 0) {
				g_warning ("Unable to create %s", to_check);
				g_free (to_check);
				return FALSE;
			}
		}

		so_far = to_check;
	}
	g_free (so_far);

	g_strfreev (split_path);

	return TRUE;
}

gboolean
kiba_utils_copy_file(const gchar *from_uri,
		const gchar *to_uri)
{
	gchar *content = NULL;
	
	if (!g_file_get_contents (from_uri, &content, NULL, NULL)) {
		return FALSE;
	}
	
	if (!g_file_set_contents (to_uri, content, -1, NULL)) {
		g_free(content);
		return FALSE;
	}
	
	g_free(content);
	return TRUE;
}

/**
 * kiba_value_array_init:
 * @num_arguments: Number of parameter, not arguments!
 * @,,,: Variable list of arguments of the form: [type1,value1,type2,value2...]
 *
 * Initiall a GValueArray for @...
 *
 * Returns: A newly allocated GValueArray.
 */
GValueArray*
kiba_utils_value_array_init (gint num_param, ...)
{
	GValueArray *array;
	GType last_type = -1;
	va_list args;
	gint num_arguments;
	gint i;

	num_arguments = num_param*2;
	array = g_value_array_new(num_arguments);

	va_start(args, num_param);
	for (i = 0; i < num_arguments; i++) {
		if (i % 2 == 0)
			last_type = va_arg (args, GType);
		else {
			GValue value = {0, };
			g_value_init(&value, last_type);
			switch (last_type) {
				case G_TYPE_BOOLEAN:
					g_value_set_boolean (&value, (gboolean) va_arg (args, int));
					break;
				case G_TYPE_CHAR:
					g_value_set_char (&value, (gchar) va_arg (args, int));
					break;
				case G_TYPE_INT:
					g_value_set_int (&value, (gint) va_arg (args, int));
					break;
				case G_TYPE_LONG:
					g_value_set_long (&value, (glong) va_arg (args, long));
					break;
				case G_TYPE_FLOAT:
					g_value_set_float  (&value, (gfloat) va_arg (args, double));
					break;
				case G_TYPE_DOUBLE:
					g_value_set_double (&value, (gdouble) va_arg (args, double));
					break;
				case G_TYPE_STRING:
					g_value_set_string (&value, (gchar*) va_arg (args, char*));
					break;
				case G_TYPE_POINTER:
					g_value_set_pointer(&value, (gpointer) va_arg (args, void*));
					break;
				case G_TYPE_UCHAR:
					g_value_set_uchar (&value, (guchar) va_arg (args, unsigned int));
					break;
				case G_TYPE_UINT:
					g_value_set_uint (&value, (guint) va_arg (args, unsigned int));
					break;
				case G_TYPE_ULONG:
					g_value_set_ulong (&value, (gulong) va_arg (args, unsigned long));
					break;
				case G_TYPE_INT64:
					g_value_set_int64 (&value, (gint64) va_arg (args, signed long));
					break;
				case G_TYPE_UINT64:
					g_value_set_uint64 (&value, (guint64) va_arg (args, long));
					break;
				default:
					g_assert_not_reached();
					break;
			}
			g_value_array_append (array, &value);
		}
	}
	va_end(args);

	return array;
}

/**
 * kiba_launch_programm:
 * @command: The command to launch
 *
 * Launch a command.
 *
 * Returns: the PID if the command was launched succesfully, otherwise 0
 */
gint
kiba_launch_programm (const gchar *command)
{
	gint argc;
	gchar **argv, **p, **q;
	GError *error = NULL;
	gint pid;

	if (!g_shell_parse_argv (command, &argc, &argv, &error)) {
		kiba_dialog_error("Failed to launch Command",
			"Failed to parse '%s': %s\n", command, error->message);
		g_error_free (error);
		return 0;
	}

	for (p = argv, q = argv; *p != NULL; p++) {
		if (*p[0] == '%' && strlen (*p) == 2) {
			switch (*p[1]) {
				case 'U':
				case 'u':
				case 'F':
				case 'f':
				default:
					g_free (*p);
					break;
			}
		}
		else
			*q++ = *p;
	}
	*q++ = NULL;
	
	g_spawn_async (NULL, argv, NULL, G_SPAWN_SEARCH_PATH,
			NULL, NULL, &pid, &error);
	if (error != NULL || pid == 0) {
		if (error != NULL) {
			kiba_dialog_error("Failed to launch Command", "%s", error->message);
			g_error_free (error);
		} else {
			kiba_dialog_error("Failed to launch Command", "Unknown Error");
		}
		return 0;
	}

	return pid;
}

/**
 * kiba_get_command_from_pid:
 * @pid: The PID.
 *
 * Trys to get a command from a PID.
 *
 * Returns: The command, or NULL.
 */
gchar*
kiba_get_command_from_pid (gint pid)
{
	GDir *proc_dir;
	gchar *buf, *pid_string;
	GError *error = NULL;

	if (pid == 0) {
		return NULL;
	}

	if (!g_file_test (PROC_PATH, G_FILE_TEST_IS_DIR)) {
		fprintf (stderr, "Cant open '" PROC_PATH "': Directory does not exist.\n");
		return NULL;
	}

	proc_dir = g_dir_open (PROC_PATH, 0, &error);
	if (error != NULL) {
		return NULL;
	}

	pid_string = g_strdup_printf ("%i", pid);

	while ((buf = (gchar *) g_dir_read_name (proc_dir)) != NULL) {
		if (g_strrstr (buf, pid_string)) {
			gsize file_length = 0;
			gchar *puffer, *config_path;

			config_path = g_strdup_printf (PROC_PATH "/%s" COMMAND_FILE, buf);

			// load file into puffer
			g_file_get_contents (config_path, &puffer, &file_length, &error);
			if (error != NULL) {
				fprintf (stderr, "Unable to load file '%s': %s",
						config_path, error->message);
				g_error_free (error);
				g_free (config_path);
				if (puffer)
					g_free (puffer);
				return  NULL;
			}

			g_dir_close (proc_dir);
			g_free (pid_string);

			return puffer;
		}
	}

	g_dir_close (proc_dir);
	g_free (pid_string);

	return NULL;
}

static gchar*
get_cpuinfo_token(gchar **info, gchar *name)
{
	gchar *line;
	*info = strstr(*info, name);
	if (*info == NULL) {
		return NULL;
	}
	*info = strstr(*info, ":");
	if (info == NULL) {
		return NULL;
	}
	*info += 2;
	line = strstr(*info, "\n");
	return g_strndup (*info, line - *info);
}

/**
 * kiba_get_cpu_frequency:
 * @cpu: cpu number.
 * 
 * Returns: CPU frequency
 */
guint
kiba_get_cpu_frequency(guint cpu)
{
	gchar *info, *buf;
	GError *error = NULL;
	gsize file_length = 0;
	guint rv = 0;

	if (!g_file_test (PROC_PATH, G_FILE_TEST_IS_DIR)) {
		fprintf (stderr, "Cant open " PROC_PATH "\n");
		return rv;
	}

	// load info
	g_file_get_contents (CPU_INFO_PATH, &info, &file_length, &error);
	if (error != NULL) {
		g_message ("Cant load " CPU_INFO_PATH
			" into buffer\nnotifications: %s", error->message);
		g_error_free (error);
		if (info) {
			g_free (info);
		}
		return rv;
	}

	// search the desired cpu
	for(buf = info;;) {
		char *token = get_cpuinfo_token(&buf, "processor");
		guint cur_cpu;
		if (token == NULL) {
			break;
		}
		cur_cpu = (guint)strtoul(token, NULL, 0);
		g_free(token);

		if (cpu == cur_cpu) {
			// found the cpu, try to get the value
			token = get_cpuinfo_token(&buf, "cpu MHz");
			if (token) {
				rv = (guint)strtoul(token, NULL, 0);
				g_free(token);
			}
			break;
		}
	}
	g_free(info);

	return rv;
}

/// GList's

/**
 * kiba_utils_glist_scroll:
 * @list: #GList to scroll.
 * @direction: scroll direction.
 *
 * "Scrolls" a Glist, moving all elements to the previous or next position.
 */
void
kiba_utils_glist_scroll (GList **list,
		GdkScrollDirection direction)
{
	GList *first, *last;

	first = *list;
	last = g_list_last (*list);

	if (direction == GDK_SCROLL_DOWN) {
		*list = last;
		first->prev = last;
		last->prev->next = NULL;
		last->prev = NULL;
		last->next = first;
	} else {
		*list = first->next;
		(*list)->prev = NULL;
		last->next = first;
		first->prev = last;
		first->next = NULL;
	}
}

/**
 * kiba_utils_sort_string_list:
 * @apointer: a listmember to compare.
 * @bpointer: the other listmember to compare.
 *
 * Gets wether apointer is alphabetical before bpointer. Use with g_list_sort on string lists.
 * 
 * Returns: a negative number if @apointer is alphabeticly before @bpointer, otherwhise a negative number.
 */
gint
kiba_utils_sort_string_list (gconstpointer apointer,
		gconstpointer bpointer)
{
	const gchar *achar = apointer;
	const gchar *bchar = bpointer;
	gint max_count;
	gint i;
	
	if (!achar) {
		if (!bchar) {
			return 0;
		} else {
			return 1;
		}
	}
	
	max_count = MIN(strlen (achar), strlen (bchar));
	
	for (i = 0; i < max_count; i++) {
		if (achar[i] > bchar[i]) {
			return 1;
		}
		if (achar[i] < bchar[i]) {
			return -1;
		}
	}

	return 0;
}

gchar*
kiba_utils_glist_to_string (GList *list)
{
	gchar *ret = NULL;
	GList *l;

	for (l = list; l != NULL; l = l->next) {
		gchar *string = l->data;

		if (ret == NULL)
			ret = g_strdup(string);
		else {
			gchar *buf = ret;
			ret = g_strdup_printf ("%s, %s", ret, string);
			g_free(buf);
		}
	}

	return ret;
}

/// String utilities

/**
 * kiba_strdown:
 * @string: string to modify.
 *
 * Replaces BIG letters with small letters.
 */
void kiba_strdown (gchar **string) {
	gint i;

	if (!(*string))
		return;

	for (i = 0; i < strlen (*string); i++) {
		(*string)[i] = g_ascii_tolower ((*string)[i]);
	}
}

/// Icon loading

gchar*
kiba_get_uri_for_icon(const gchar *icon,
		gint flags,
		gint size)
{
	GtkIconInfo *icon_info;
	GError *error = NULL;
	gchar *path, *suffix;

	g_return_val_if_fail (icon != NULL, FALSE);

	if (g_file_test (icon, G_FILE_TEST_IS_REGULAR)) {
		return g_strdup(icon);
	}

	// try icon theme
	icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default(),
		icon, size, flags);
	if (icon_info != NULL) {
		gchar *path = g_strdup(gtk_icon_info_get_filename (icon_info));
		gtk_icon_info_free (icon_info);
		return path;
	}

	// try png at /usr/share/pixmaps path
	path = g_strdup_printf (PIXMAP_PATH "/%s", icon);
	if (g_file_test (path, G_FILE_TEST_IS_REGULAR)) {
		return path;
	}
	g_free (path);

	// try png at
	g_clear_error (&error);
	path = g_strdup(icon);
	suffix = strstr (path, ".png");
	if (suffix == NULL) {
		suffix = strstr (path, ".svg");
	}
	if (suffix == NULL) {
		return NULL;
	}
	*suffix = '\0';
	icon_info = gtk_icon_theme_lookup_icon (
		gtk_icon_theme_get_default (),
		path, size, flags);
	g_free(path);
	if (icon_info != NULL) {
		path = g_strdup(gtk_icon_info_get_filename (icon_info));
		gtk_icon_info_free (icon_info);
		return path;
	}

	return NULL;
}

/**
 * kiba_load_svg:
 * @icon: icon path or icon name.
 * @handle: #RsvgHandle return.
 * @dimension_data: handle size ret.
 *
 * Trys to load a svg file.
 *
 * Returns: TRUE if the svg was loaded succesfully, otherwise FALSE is returned.
 */
gboolean
kiba_load_svg (const gchar *icon,
		RsvgHandle **handle,
		RsvgDimensionData *dimension_data)
{
	GtkIconInfo *icon_info = NULL;
	gchar *path;

	g_return_val_if_fail (icon != NULL, FALSE);
	if (g_str_has_suffix (icon, ".png")) {
		return FALSE;
	}
	if (g_str_has_suffix (icon, ".xpm")) {
		return FALSE;
	}

	path = kiba_get_uri_for_icon(icon,
		GTK_ICON_LOOKUP_FORCE_SVG,
		0);
	if (!g_str_has_suffix (path, ".svg")) {
		g_free(path);
		// FIXME: some icons (utility-terminal) load with 48 set here, others not :/
		path = kiba_get_uri_for_icon(icon,
			GTK_ICON_LOOKUP_FORCE_SVG,
			48);
	}
	if (!path) {
		DEBUG_WARNING ("Failed to get path for '%s'\n", icon);
		return FALSE;
	}

	if (*handle != NULL) {
		rsvg_handle_free (*handle);
		*handle = NULL;
	}

	// try SVG with absolute path
	*handle = rsvg_handle_new_from_file (path, NULL);
	if (*handle) {
		if (dimension_data != NULL)
			rsvg_handle_get_dimensions (*handle, dimension_data);
		return TRUE;
	}

	DEBUG_WARNING ("Failed to load SVG for '%s'\n", path);

	return FALSE;
}

/**
 * kiba_load_icon:
 * @icon: The icon path or icon name.
 * @size: The pixbuf size.
 * @pixbuf: The GdkPixbuf return.
 *
 * Loads a icon file.
 *
 * Returns: TRUE if the icon was loaded succesfully, otherwise FALSE is returned.
 */
gboolean
kiba_load_icon (const gchar *icon,
		gint size,
		GdkPixbuf **pixbuf)
{
	GtkIconInfo *icon_info;
	GError *error = NULL;
	gchar *path;

	g_return_val_if_fail (icon != NULL, FALSE);
	g_return_val_if_fail (size > 0, FALSE);

	path = kiba_get_uri_for_icon(icon,
		GTK_ICON_LOOKUP_NO_SVG, size);
	if (!path) {
		DEBUG_WARNING ("Failed to get path for '%s'\n", icon);
		return FALSE;
	}

	if (*pixbuf != NULL) {
		g_object_unref (*pixbuf);
		*pixbuf = NULL;
	}

	// try png with absolute path
	*pixbuf = gdk_pixbuf_new_from_file_at_size (path, size, -1, &error);
	if (*pixbuf != NULL) {
		g_free(path);
		return TRUE;
	}

	if (error != NULL) {
		DEBUG_WARNING ("Failed to load Pixbuf for '%s': %s\n", path, error->message);
		g_error_free (error);
	} else {
		DEBUG_WARNING ("Failed to load Pixbuf for '%s'\n", path);
	}
	g_free(path);

	return FALSE;
}
