/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <daniel@orgizm.net>
*/

#include <librsvg/rsvg.h>
#include <cairo.h>
#include <math.h>

#include "config.h"

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "activity.h"

#define ACTIVITY_FEATURE_ICON DATADIR"/kiba-dock/icons/broken-glass.svg"

G_DEFINE_TYPE (KibaActivityFeature, kiba_activity_feature, KIBA_TYPE_VISUAL_FEATURE)

/// Private methods

static void
resize_feature_object(
		KibaVisualFeature *feature,
		KibaObject *object)
{
	
	GList *l;
	KibaVisualFeatureObject *vobject = NULL;
	for (l = feature->objects; l != NULL; l = l->next) {
		vobject = l->data;
		if (vobject->object == object) break;
	}
	if (vobject) {
		kiba_visual_feature_queue_resize(feature, vobject);
	}
}

static void
__scale_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}

static void
__offset_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}

static void
__activity_toggled(KibaObject *object,
		gboolean active,
		gpointer data)
{
	if (active) {
		kiba_visual_feature_add_object(
			KIBA_VISUAL_FEATURE(data), object);
	} else {
		kiba_visual_feature_remove_object(
			KIBA_VISUAL_FEATURE(data), object);
	}
}

static void
__draw_surface (cairo_t *cr,
		cairo_surface_t  *surface,
		gint width,
		gint height,
		void *data)
{
	GdkPixbuf *pixbuf;
	RsvgHandle *handle = NULL;
	RsvgDimensionData dimension_data;

	// TODO: use gboolean kiba_surface_factory_load_icon (icon, size, func, data)

	if (kiba_load_svg (ACTIVITY_FEATURE_ICON, &handle, &dimension_data)) {
		cairo_scale(cr,
			(gfloat) width/dimension_data.width,
			(gfloat) height/dimension_data.height);
		rsvg_handle_render_cairo (handle, cr);
		rsvg_handle_free (handle);
		return;
	}
	pixbuf = NULL;
	if (kiba_load_icon (ACTIVITY_FEATURE_ICON, width, &pixbuf)) {
		gdk_cairo_set_source_pixbuf (cr, pixbuf, 0.0, 0.0);
		cairo_paint (cr);
		g_object_unref (pixbuf);
		return;
	}

	// fallback to cairo drawing

	cairo_set_line_width (cr, 2.0);
	cairo_set_line_join  (cr, CAIRO_LINE_JOIN_ROUND);

	cairo_move_to (cr, 0.5*width, 0.1*height);
	cairo_line_to (cr, 0.9*width, 0.7*height);
	cairo_line_to (cr, 0.1*width, 0.7*height);
	cairo_line_to (cr, 0.5*width, 0.1*height);
	cairo_close_path (cr);

	cairo_set_source_rgba (cr, 1, 1, 1, 0.8);
	cairo_fill_preserve (cr);
	cairo_set_source_rgba (cr, 0.8, 0.8, 0.8, 0.6);
	cairo_stroke (cr);
}

static void
__update_surface (KibaActivityFeature *activity, KibaWin *win)
{
	cairo_t *cr = kiba_win_cairo_create (win);
	cairo_update_surface_at_size (cr, &activity->surface,
		ACTIVITY_FEATURE_SIZE, ACTIVITY_FEATURE_SIZE,
		__draw_surface, NULL);
	cairo_destroy (cr);
}

/// Class implementation

static void
kiba_activity_feature_object_added(KibaVisualFeature *feature,
		KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object)) {
		return;
	}
	
	g_signal_connect(G_OBJECT(object), "scale-changed",
		G_CALLBACK(__scale_changed_cb), feature);
	g_signal_connect(G_OBJECT(object), "offset-changed",
		G_CALLBACK(__offset_changed_cb), feature);
	g_signal_connect(G_OBJECT(object), "active-toggled",
		G_CALLBACK(__activity_toggled), feature);
	if (kiba_object_get_state(object, KIBA_OBJECT_ACTIVE)) {
		kiba_visual_feature_add_object(feature, object);
	}
}

static void
kiba_activity_feature_object_removed(KibaVisualFeature *feature,
		KibaObject *object)
{
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__activity_toggled), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__scale_changed_cb), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__offset_changed_cb), feature);
	if (!kiba_object_get_state(object, KIBA_OBJECT_ACTIVE)) {
		kiba_visual_feature_remove_object(feature, object);
	}
}

static void
kiba_activity_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaActivityFeature *activity = KIBA_ACTIVITY_FEATURE(feature);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	const KibaLayoutData *d;
	gfloat alpha = 
		KIBA_DRAWABLE_OBJECT(object->object)->alpha;
	
	d = &kiba_layout_data[dock->d.position];
	
	if (alpha <= 0.0f) {
		return;
	}
	
	cairo_translate (cr,
		object->allocation.width/2,
		object->allocation.height/2);
	if (d->anchor.y == 0) {
		cairo_rotate (cr, M_PI);
	} else if (d->anchor.x == 0) {
		cairo_rotate (cr, M_PI/2);
	} else if (d->anchor.x == 4) {
		cairo_rotate (cr, 3*M_PI/2);
	}
	cairo_translate (cr,
		-object->allocation.width/2,
		-object->allocation.height/2);

	cairo_set_source_surface (cr, activity->surface, 0, 0);
	cairo_paint_with_alpha (cr, alpha);
}

static KibaAllocation
kiba_activity_feature_get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	GdkRectangle area = kiba_drawable_object_get_area (drawable_object);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	KibaAllocation alloc;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[dock->d.position];

	alloc = (KibaAllocation) {
		0.0f, 0.0f,
		ACTIVITY_FEATURE_SIZE,
		ACTIVITY_FEATURE_SIZE
	};
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		alloc.y = 0.5f*(object->object->alloc.height - ACTIVITY_FEATURE_SIZE);
		alloc.x = ((gfloat)d->halignment-1.0f)*(
			object->object->alloc.width - 10.0f) -
			(gfloat)((d->halignment-2)/2)*(
				object->object->alloc.width - ACTIVITY_FEATURE_SIZE);
	} else {
		alloc.x = 0.5f*(object->object->alloc.width - ACTIVITY_FEATURE_SIZE);
		alloc.y = ((gfloat)d->valignment-1.0f)*(
			object->object->alloc.height - 10.0f) -
			(gfloat)((d->valignment-2)/2)*(
				object->object->alloc.height - ACTIVITY_FEATURE_SIZE);
	}
	
	return alloc;
}

static void
kiba_activity_feature_win_set(KibaVisualFeature *feature, KibaWin *win)
{
	KibaActivityFeature *activity = KIBA_ACTIVITY_FEATURE(feature);
	__update_surface(activity, win);
}

static void
kiba_activity_feature_class_init (KibaActivityFeatureClass *class)
{
}

static void
kiba_activity_feature_init (KibaActivityFeature *activity)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(activity);
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_activity_feature_get_allocation;
	feature->draw = kiba_activity_feature_draw;
	feature->object_added = kiba_activity_feature_object_added;
	feature->object_removed = kiba_activity_feature_object_removed;
	feature->win_set = kiba_activity_feature_win_set;
}

KibaVisualFeature*
kiba_activity_feature_new(KibaWin *win)
{
	KibaVisualFeature *feature = g_object_new (KIBA_TYPE_ACTIVITY_FEATURE,
		    "name",  ACTIVITY_FEATURE_NAME,
		    "type",  KIBA_VISUAL_FEATURE_ONTOP_OF_DOCK,
		    "front", TRUE,
		    NULL);
	kiba_visual_feature_set_win(feature, win);
	return feature;
}
