/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <daniel@orgizm.net>
*/

#include <cairo.h>

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "prelight.h"

G_DEFINE_TYPE (KibaPrelightFeature, kiba_prelight_feature, KIBA_TYPE_VISUAL_FEATURE)

/*** Private methods */

static void
__prelight_toggled(KibaObject *object,
	gboolean prelight,
	gpointer data)
{
	if (prelight) {
		kiba_visual_feature_add_object(KIBA_VISUAL_FEATURE(data), object);
	} else {
		kiba_visual_feature_remove_object(KIBA_VISUAL_FEATURE(data), object);
	}
}

static void
__prelight_rgba_cb(const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaPrelightFeature *prelight = KIBA_PRELIGHT_FEATURE(data);
	kiba_prelight_feature_set_rgba(prelight,
		kiba_prefs_rgba_from_string ((char*) value));
}

/// Public methods

void
kiba_prelight_feature_set_rgba(KibaPrelightFeature *prelight, KibaRGBA rgba)
{
	g_return_if_fail(KIBA_IS_PRELIGHT_FEATURE(prelight));
	prelight->rgba = rgba;
}

KibaRGBA
kiba_prelight_feature_get_rgba(KibaPrelightFeature *prelight)
{
	g_return_if_fail(KIBA_IS_PRELIGHT_FEATURE(prelight));
	return prelight->rgba;
}

/// Class implementation

static void
kiba_prelight_feature_object_added(KibaVisualFeature *feature, KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object) || KIBA_IS_SEPARATOR(object)) {
		return;
	}
	
	g_signal_connect(G_OBJECT(object), "prelight-toggled",
		G_CALLBACK(__prelight_toggled), feature);
	if (kiba_object_get_state(object, KIBA_OBJECT_PRELIGHTED)) {
		kiba_visual_feature_add_object(feature, object);
	}
}

static void
kiba_prelight_feature_object_removed(KibaVisualFeature *feature, KibaObject *object)
{
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__prelight_toggled), feature);
	if (kiba_object_get_state(object, KIBA_OBJECT_PRELIGHTED))
		kiba_visual_feature_remove_object(feature, object);
}

static void
kiba_prelight_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaPrelightFeature *prelight = KIBA_PRELIGHT_FEATURE(feature);
	gfloat alpha = KIBA_DRAWABLE_OBJECT(
			object->object)->alpha * prelight->rgba.a;
	if (alpha <= 0.0f) return;
	cairo_set_source_rgba (cr,
		prelight->rgba.r,
		prelight->rgba.g,
		prelight->rgba.b,
		alpha);
	cairo_paint (cr);
}

static KibaAllocation
kiba_prelight_feature_get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	GdkRectangle area = kiba_drawable_object_get_area (
		KIBA_DRAWABLE_OBJECT(object->object));
	return (KibaAllocation) {
		0.0f, 0.0f,
		area.width,
		area.height
	};
}

static void
kiba_prelight_feature_class_init (KibaPrelightFeatureClass *class)
{
}

static void
kiba_prelight_feature_init (KibaPrelightFeature *prelight)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(prelight);
	
	prelight->rgba = (KibaRGBA) { 1.0f, 1.0f, 1.0f, 0.4f };
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_prelight_feature_get_allocation;
	feature->draw = kiba_prelight_feature_draw;
	feature->object_added = kiba_prelight_feature_object_added;
	feature->object_removed = kiba_prelight_feature_object_removed;
}

KibaVisualFeature*
kiba_prelight_feature_new(KibaWin *win, KibaPrefsFile *file)
{
	KibaVisualFeature *feature = g_object_new (KIBA_TYPE_PRELIGHT_FEATURE,
		    "name",  PRELIGHT_FEATURE_NAME,
		    "type",  KIBA_VISUAL_FEATURE_ONTOP,
		    "front", TRUE,
		    NULL);
	kiba_visual_feature_set_win(feature, win);

	kiba_prefs_add_options (file,
	    (PrefMember[]) { { "prelight_rgba", "#FFFFFF66", KIBA_STRING } },
	    1
	);
	kiba_prefs_add_callback (file, "prelight_rgba",
		__prelight_rgba_cb, feature, TRUE);
	
	return feature;
}
