/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <gtk/gtk.h>
#include <librsvg/rsvg.h>
#include <cairo.h>
#include <math.h>

#include "progress.h"
#include "kiba-macros.h"

#define PROGRESS_DATA_KEY  "progress-object"
#define FADE_STEP           0.065f
#define FADE_RATE           75
#define ICON_PATH           KIBA_ICON_PATH"/progress.svg"
#define PATTERN_ICON_PATH   KIBA_ICON_PATH"/progress-pattern.svg"

G_DEFINE_TYPE (KibaProgressFeature, kiba_progress_feature, KIBA_TYPE_VISUAL_FEATURE)

/*** Private methods */

static KibaProgressFeatureObject* __find_pobject(GList *objects, gint id)
{
	GList *l;
	for (l = objects; l != NULL; l = l->next) {
		KibaProgressFeatureObject *obj = l->data;
		if (obj->id == id) return obj;
	}
	return NULL;
}

static gint __fade(gpointer data)
{
	KibaProgressFeatureObject *pobject = data;
	
	if (pobject->running) {
		pobject->alpha = MIN(1.0f, pobject->alpha+FADE_STEP);
		kiba_visual_feature_queue_draw(pobject->feature, pobject->object);
		if (pobject->alpha == 1.0f) {
			pobject->timeout_id = 0;
			return FALSE;
		} else {
			return TRUE;
		}
	} else {
		pobject->alpha = MAX(0.0f, pobject->alpha-FADE_STEP);
		kiba_visual_feature_queue_draw(pobject->feature, pobject->object);
		if (pobject->alpha == 0.0f) {
			GList *processes = g_object_get_data(
				G_OBJECT(pobject->object->object), PROGRESS_DATA_KEY);
			pobject->timeout_id = 0;
			// remove object now
			processes = g_list_remove(processes, pobject);
			g_object_set_data(G_OBJECT(pobject->object->object),
				PROGRESS_DATA_KEY, processes);
			kiba_visual_feature_remove_object(pobject->feature,
				pobject->object->object);
			g_free(pobject);
			return FALSE;
		} else {
			return TRUE;
		}
	}
}

static void __process_toggled_cb(KibaObject *object,
	gint id,
	gboolean started,
	gpointer data)
{
	KibaVisualFeature *feature = data;
	KibaProgressFeatureObject *pobject;
	GList *processes = g_object_get_data(
		G_OBJECT(object), PROGRESS_DATA_KEY);
	
	if (started) {
		pobject = g_new0(KibaProgressFeatureObject, 1);
		processes = g_list_prepend(processes, pobject);
		g_object_set_data(
			G_OBJECT(object), PROGRESS_DATA_KEY, processes);
		
		pobject->feature = feature;
		pobject->alpha = 0.0f;
		pobject->value = 0;
		pobject->id = id;
		pobject->running = TRUE;
		pobject->timeout_id = g_timeout_add(FADE_RATE, __fade, pobject);
		pobject->object = kiba_visual_feature_add_object(KIBA_VISUAL_FEATURE(data), object);
	} else {
		pobject = __find_pobject(processes, id);
		if (pobject != NULL) {
			pobject->running = FALSE;
			if (!pobject->timeout_id)
				pobject->timeout_id = g_timeout_add(FADE_RATE, __fade, pobject);
		}
	}
}

static void __process_updated_cb(KibaObject *object,
	gint id,
	gint value,
	gpointer data)
{
	KibaProgressFeatureObject *pobject;
	GList *processes = g_object_get_data(
		G_OBJECT(object), PROGRESS_DATA_KEY);
	
	pobject = __find_pobject(processes, id);
	g_return_if_fail(pobject != NULL);
	
	pobject->value = value;
	kiba_visual_feature_queue_draw(pobject->feature, pobject->object);
}

static void __draw_progress_pattern (cairo_t      *cr,
			cairo_surface_t  *surface,
			gint              width,
			gint              height,
			void             *data) {
	KibaProgressFeature *progress =
		KIBA_PROGRESS_FEATURE(data);
	RsvgHandle *handle = NULL;
	RsvgDimensionData dimension_data;
	if (kiba_load_svg (PATTERN_ICON_PATH, &handle, &dimension_data)) {
		cairo_scale(cr,
			(float) width/dimension_data.width,
			(float) height/dimension_data.height);
		rsvg_handle_render_cairo (handle, cr);
		rsvg_handle_free (handle);
	}
}

static void __draw_progress_surface (cairo_t      *cr,
			cairo_surface_t  *surface,
			gint              width,
			gint              height,
			void             *data) {
	cairo_pattern_t *pattern;
	cairo_surface_t *pattern_surface = NULL;
	KibaProgressFeature *progress =
		KIBA_PROGRESS_FEATURE(data);
	
	// draw the pattern image to a surface
	cairo_update_surface_at_size (cr,
		&pattern_surface,
		width, 3.0f,
		__draw_progress_pattern,
		progress);
	
	// create the pattern
	pattern = cairo_pattern_create_for_surface (pattern_surface);
	cairo_surface_destroy (pattern_surface);
#if CAIRO_VERSION_MAJOR < 1 || (CAIRO_VERSION_MAJOR == 1 && CAIRO_VERSION_MINOR < 4)
	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
#else
	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REFLECT);
#endif
	
	// set pattern as source
	cairo_set_source (cr, pattern);
	// draw the pattern using the alphamask from progress surface
	cairo_mask_surface (cr, progress->surface, 0, 0);
	// cleanup
	cairo_pattern_destroy (pattern);
}

static void __draw_overlay_surface (cairo_t      *cr,
			cairo_surface_t  *surface,
			gint              width,
			gint              height,
			void             *data) {
	KibaProgressFeature *progress =
		KIBA_PROGRESS_FEATURE(data);
	RsvgDimensionData dimension_data;
	RsvgHandle *handle = NULL;
	if (kiba_load_svg (ICON_PATH, &handle, &dimension_data)) {
		cairo_scale(cr,
			(float) width/dimension_data.width,
			(float) height/dimension_data.height);
		rsvg_handle_render_cairo_sub (handle, cr, "#layer2");
		rsvg_handle_free (handle);
	}
}

static void __draw_surface (cairo_t      *cr,
			cairo_surface_t  *surface,
			gint              width,
			gint              height,
			void             *data) {
	KibaProgressFeature *progress =
		KIBA_PROGRESS_FEATURE(data);
	RsvgDimensionData dimension_data;
	RsvgHandle *handle = NULL;
	if (kiba_load_svg (ICON_PATH, &handle, &dimension_data)) {
		cairo_scale(cr,
			(float) width/dimension_data.width,
			(float) height/dimension_data.height);
		rsvg_handle_render_cairo_sub (handle, cr, "#layer1");
		rsvg_handle_free (handle);
	}
}

static void __update_surface (KibaProgressFeature *progress, KibaWin *win) {
	cairo_t *cr;
	
	if (progress->size.height == 0.0f || progress->size.width == 0.0f)
		return;
	
	cr = kiba_win_cairo_create (win);
	if (!cr) return;
	
	cairo_update_surface_at_size (cr,
		&progress->surface,
		progress->size.width, progress->size.height,
		__draw_surface,
		progress);
	cairo_update_surface_at_size (cr,
		&progress->overlay_surface,
		progress->size.width, progress->size.height,
		__draw_overlay_surface,
		progress);
	cairo_update_surface_at_size (cr,
		&progress->progress_surface,
		progress->size.width, progress->size.height,
		__draw_progress_surface,
		progress);
	cairo_destroy (cr);
}


static void __object_size_changed (KibaDock *dock,
		gint new_size,
		gpointer data)
{
	KibaProgressFeature *progress = data;
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(progress);
	
	progress->size.width = new_size*0.425f;
	progress->size.height = new_size*1.1f;
	__update_surface(progress, KIBA_WIN(GTK_WIDGET(dock)->parent));
	
	kiba_visual_feature_queue_resize_all(feature);
}

/*** Public methods */


/*** Class implementation */

static void kiba_progress_feature_object_added(KibaVisualFeature *feature, KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object))
		return;
	
	GList *l;
	
	g_signal_connect(G_OBJECT(object), "process-toggled",
		G_CALLBACK(__process_toggled_cb), feature);
	g_signal_connect(G_OBJECT(object), "process-updated",
		G_CALLBACK(__process_updated_cb), feature);
	
	for (l = kiba_object_get_processes(object); l != NULL; l = l->next) {
		KibaObjectProgress *progress = l->data;
		__process_toggled_cb(object, progress->id, TRUE, feature);
		__process_updated_cb(object, progress->id, progress->value, feature);
	}
}

static void kiba_progress_feature_object_removed(KibaVisualFeature *feature, KibaObject *object)
{
	GList *l;
	GList *processes = g_object_get_data(
		G_OBJECT(object), PROGRESS_DATA_KEY);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__process_toggled_cb), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__process_updated_cb), feature);
	
	for (l = kiba_object_get_processes(object); l != NULL; l = l->next) {
		KibaObjectProgress *progress = l->data;
		KibaProgressFeatureObject *pobject =
			__find_pobject(processes, progress->id);
		if (pobject->timeout_id)
			g_source_remove(pobject->timeout_id);
		g_free(pobject);
	}
}

static void kiba_progress_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaProgressFeature *progress =
		KIBA_PROGRESS_FEATURE(feature);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	GList *processes = g_object_get_data(
		G_OBJECT(object->object), PROGRESS_DATA_KEY);
	const KibaLayoutData *d;
	float alpha;
	// FIXME: support multiple progresses
	KibaProgressFeatureObject *pobject = processes->data;
	
	d = &kiba_layout_data[dock->d.position];
	alpha = pobject->alpha*KIBA_DRAWABLE_OBJECT(object->object)->alpha;
	
	if (alpha <= 0.0f || object->object->widget_scale_x == 0.0f || object->object->widget_scale_y == 0.0f) {
		return;
	}
	
	// draw background
	cairo_save(cr);
		cairo_scale(cr, object->object->widget_scale_x, object->object->widget_scale_y);
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			cairo_rotate(cr, M_PI/2.0f);
			cairo_translate(cr, 0, -object->allocation.width);
		}
		cairo_set_source_surface (cr, progress->surface, 0, 0);
		cairo_paint_with_alpha(cr, alpha);
	cairo_restore(cr);
	
	// draw progress
	cairo_save(cr);
		float percent = pobject->value/100.0f;
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			cairo_rectangle(cr,
				(d->halignment/2)*(1.0f-percent)*object->allocation.width,
				0,
				percent*object->allocation.width,
				object->allocation.height
			);
		} else {
			cairo_rectangle(cr,
				0,
				(d->valignment/2)*(1.0f-percent)*object->allocation.height,
				object->allocation.width,
				percent*object->allocation.height
			);
		}
		cairo_clip(cr);
		
		cairo_scale(cr, object->object->widget_scale_x, object->object->widget_scale_y);
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			cairo_rotate(cr, M_PI/2.0f);
			cairo_translate(cr, 0, -object->allocation.width);
		}
		cairo_set_source_surface (cr, progress->progress_surface, 0, 0);
		cairo_paint_with_alpha(cr, alpha);
	cairo_restore(cr);
	
	// draw border
	cairo_save(cr);
		cairo_scale(cr, object->object->widget_scale_x, object->object->widget_scale_y);
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			cairo_rotate(cr, M_PI/2.0f);
			cairo_translate(cr, 0, -object->allocation.width);
		}
		cairo_set_source_surface (cr, progress->overlay_surface, 0, 0);
		cairo_paint_with_alpha(cr, alpha);
	cairo_restore(cr);
}

static KibaAllocation kiba_progress_feature_get_allocation(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	GdkRectangle area = kiba_drawable_object_get_area (drawable_object);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	KibaProgressFeature *progress = KIBA_PROGRESS_FEATURE(feature);
	KibaAllocation alloc;
	const KibaLayoutData *d;
	gfloat height, width;
	
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		height = object->object->widget_scale_y*progress->size.height;
		width = object->object->widget_scale_x*progress->size.width;
	} else {
		width = object->object->widget_scale_x*progress->size.height;
		height = object->object->widget_scale_y*progress->size.width;
	}

	alloc = (KibaAllocation) { 0.0f, 0.0f, width, height };
	
	// translate against offset
	alloc.x = ((float)d->halignment-1.0f)*object->object->surface_offset_x;
	alloc.y = ((float)d->valignment-1.0f)*object->object->surface_offset_y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		alloc.y += ((d->valignment-2)/2)*(height-object->object->alloc.height);
		alloc.x += object->object->alloc.width - alloc.width;
	} else {
		alloc.x += ((d->halignment-2)/2)*(width-object->object->alloc.width);
		alloc.y += object->object->alloc.height - alloc.height;
	}
	
	return alloc;
}

static void kiba_progress_feature_win_set(KibaVisualFeature *feature, KibaWin *win)
{
	KibaProgressFeature *progress =
		KIBA_PROGRESS_FEATURE(feature);
	KibaDock *dock = kiba_visual_feature_get_dock(feature);

	__update_surface(progress, win);
	
	__object_size_changed(dock, dock->d.object_size, progress);
	g_signal_connect(G_OBJECT(dock), "object-size-changed",
		G_CALLBACK(__object_size_changed), progress);
}

static void kiba_progress_feature_dispose(GObject *object)
{
	KibaProgressFeature *progress =
		KIBA_PROGRESS_FEATURE(object);
}

static void kiba_progress_feature_class_init (KibaProgressFeatureClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	
	// GObject overrides
	gobject_class->dispose = kiba_progress_feature_dispose;
}

static void kiba_progress_feature_init (KibaProgressFeature *progress)
{
	KibaVisualFeature *feature = KIBA_VISUAL_FEATURE(progress);
	
	progress->size.width = 0.0f;
	progress->size.height = 0.0f;
	progress->surface = NULL;
	
	// KibaVisualFeature overrides
	feature->get_allocation = kiba_progress_feature_get_allocation;
	feature->draw = kiba_progress_feature_draw;
	feature->object_added = kiba_progress_feature_object_added;
	feature->object_removed = kiba_progress_feature_object_removed;
	feature->win_set = kiba_progress_feature_win_set;
}

KibaVisualFeature* kiba_progress_feature_new(KibaWin *win)
{
	KibaVisualFeature *feature = g_object_new (
		    KIBA_TYPE_PROGRESS_FEATURE,
		    "name",  ABSOLUTE_PROGRESS_FEATURE_NAME,
		    "type",  KIBA_VISUAL_FEATURE_GIMMICK,
		    "front", TRUE,
		    NULL);
	kiba_visual_feature_set_win(feature, win);
	
	return feature;
}
