/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"

#include <string.h>
#include <pthread.h>
#include <gtk/gtk.h>

#include "kiba-utils.h"
#include "kiba-icon-factory.h"

struct _KibaIconFactoryIconInfo {
	GList *callbacks;
	union {
		GdkPixbuf *pixbuf;
		RsvgHandle *svg;
	}icon;
	gint size;
	gint reference_count;
	gchar *uri;
};

typedef struct {
	KibaIconFactoryCallback func;
	gpointer user_data;
}KibaIconFactoryCallbackData;

// List of currently used icons.
static GList *icons = NULL;

// Our mutex used when accessing data shared between the main thread and the
// icon thread, i.e. the icon_thread_is_running flag and the
// icons_to_make list.
static pthread_mutex_t icons_mutex = PTHREAD_MUTEX_INITIALIZER;
// The currently surfaceed icon. it also exists in the icons_to_make list
// to avoid adding it again. Lock icons_mutex when accessing this.
static KibaIconFactoryIconInfo *currently_loading = NULL;
// The list of KibaIconFactoryIconInfo structs containing information about the
// icons we are making. Lock icons_mutex when accessing this.
static volatile GList *icons_to_make = NULL;
// List of icons that need to invoke the callbacks.
// Lock icons_mutex when accessing this
static volatile GList *icons_to_load = NULL;
// A flag to indicate whether a icon thread is running, so we don't
// start more than one. Lock icons_mutex when accessing this. */
static volatile gboolean icon_thread_is_running = FALSE;
// timeout that handles loaded ions
static gint icon_load_timeout_id = 0;

/// Private methods

static void
destroy_icon_info (KibaIconFactoryIconInfo *info)
{
	g_list_free(info->callbacks);
	g_free(info->uri);
	g_free(info);
}

// read_icon is invoked in a separate thread.
static void
read_icon (KibaIconFactoryIconInfo *info)
{
	GdkPixbuf *pixbuf = NULL;
	GtkIconInfo *iconInfo = NULL;
	RsvgHandle *handle = NULL;
	RsvgDimensionData dimension_data;

	if (kiba_load_svg (info->uri, &handle, &dimension_data)) {
		info->icon.svg = handle;
		return;
	}
	if (kiba_load_icon (info->uri, info->size, &pixbuf)) {
		info->icon.pixbuf = pixbuf;
		return;
	}
	
	iconInfo = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default(),
		GTK_STOCK_MISSING_IMAGE, info->size, 0);
	if (iconInfo != NULL) {
		const gchar *path;

		path = gtk_icon_info_get_filename (iconInfo);
		gtk_icon_info_free(iconInfo);
		info->icon.pixbuf = gdk_pixbuf_new_from_file_at_size (
			path, info->size, -1, NULL);
	}
}

// walk through icons_to_load list and invoke callbacks, this is done in the main thread
static gint
load_icon (gpointer data)
{
	GList *l;
	
	for (l = (GList*) icons_to_load; l != NULL; l = (GList*) icons_to_load) {
		KibaIconFactoryIconInfo *info = l->data;
		GList *m;

		/// MUTEX LOCKED
		pthread_mutex_lock (&icons_mutex); {
			icons_to_load = g_list_remove ((GList*) icons_to_load, info);
		} pthread_mutex_unlock (&icons_mutex);
		/// MUTEX UNLOCKED

		icons = g_list_append(icons, info);

		for (m = info->callbacks; m != NULL; m = info->callbacks) {
			KibaIconFactoryCallbackData *fdata = m->data;
			info->callbacks = g_list_remove(info->callbacks, m->data);
			
			if (GDK_IS_PIXBUF(info->icon.pixbuf)) {
				fdata->func(info->icon.pixbuf, NULL,
					info->size, fdata->user_data);
			} else if (RSVG_IS_HANDLE(info->icon.svg)) {
				fdata->func(NULL, info->icon.svg,
					info->size, fdata->user_data);
			}
		}
	}

	if (icons_to_make != NULL) {
		// still icons to make
		return TRUE;
	}
	
	// no icons left for loading
	icon_load_timeout_id = 0;
	return FALSE;
}

// icon_thread is invoked as a separate thread to make icons.
static gpointer
thread_start (gpointer data)
{
	KibaIconFactoryIconInfo *info = NULL;
	GList *l;

	// We loop until there are no more icons to make, at which point
	// we exit the thread.
	for (;;) {
		/// MUTEX LOCKED
		pthread_mutex_lock (&icons_mutex);

		// Pop the last icon we just made off the head of the
		// list and free it. I did this here so we only have to lock
		// the mutex once per icon, rather than once before
		// creating it and once after.
		if (currently_loading) {
			g_assert (info == currently_loading);
			icons_to_make = g_list_remove ((GList*) icons_to_make, currently_loading);
		}
		currently_loading = NULL;

		// If there are no more icons to make, reset the
		// icon_thread_is_running flag, unlock the mutex, and
		// exit the thread.
		if (icons_to_make == NULL) {
			icon_thread_is_running = FALSE;
			pthread_mutex_unlock (&icons_mutex);
			pthread_exit (NULL);
		}
		pthread_mutex_unlock (&icons_mutex);
		/// MUTEX UNLOCKED

		// Get the next one to make. We leave it on the list until it
		// is created so the main thread doesn't add it again while we
		// are creating it.
		info = icons_to_make->data;
		currently_loading = info;

		// load the icon
		read_icon(info);

		/// MUTEX LOCKED
		pthread_mutex_lock (&icons_mutex); {
			icons_to_load = g_list_append((GList*) icons_to_load, info);
		} pthread_mutex_unlock (&icons_mutex);
		/// MUTEX UNLOCKED
	}
}

static KibaIconFactoryIconInfo*
find_icon_info(const gchar *icon_path)
{
	GList *l;
	for (l = (GList*) icons; l != NULL; l = l->next) {
		KibaIconFactoryIconInfo *data = l->data;
		if (strcmp(data->uri, icon_path) == 0) {
			return data;
		}
	}
	return NULL;
}

/// Public methods

/**
 * kiba_icon_factory_unload_icon:
 * @icon_info: icon information returned by kiba_surface_factory_load_icon
 *
 * Decrease reference count by one, and removes the icon from the factory
 * if the reference counter reached 0.
 */
void
kiba_icon_factory_unload_icon (KibaIconFactoryIconInfo *icon_info)
{
	g_return_if_fail(icon_info != NULL);

	icon_info->reference_count--;
	if (icon_info->reference_count > 0) {
		return;
	}
	
	// remove from icons_to_load
	pthread_mutex_lock (&icons_mutex); {
		if (g_list_find((GList*) icons_to_load, icon_info)) {
			icons_to_load = g_list_remove ((GList*) icons_to_load, icon_info);
		}
	} pthread_mutex_unlock (&icons_mutex);
	
	icons = g_list_remove(icons, icon_info);

	if (GDK_IS_PIXBUF(icon_info->icon.pixbuf)) {
		g_object_unref(G_OBJECT(icon_info->icon.pixbuf));
	} else if (RSVG_IS_HANDLE(icon_info->icon.svg)) {
		rsvg_handle_free(icon_info->icon.svg);
	}
	
	destroy_icon_info(icon_info);
}

/**
 * kiba_icon_factory_load_icon:
 * @icon: The icon to load.
 * @size: Size for @icon.
 * @func: Callback to invoke after the surface was loaded.
 * @user_data: Userdata for @func.
 *
 * Load @icon in a separate thread (if the icon isnt allready added to the factory)
 * and invoke @func in the main thread.
 * Please note that @func may is not called immediately.
 * 
 * Returns: the #KibaIconFactoryIconInfo created, or NULL on failure.
 */
KibaIconFactoryIconInfo*
kiba_icon_factory_load_icon (const gchar *icon,
		gint size,
		KibaIconFactoryCallback func,
		gpointer user_data)
{
	KibaIconFactoryCallbackData *callback_data;
	KibaIconFactoryIconInfo *data;
	gchar *icon_path;
	GList *l;
	
	g_return_if_fail(func != NULL);
	g_return_if_fail(icon != NULL);

	icon_path = kiba_get_uri_for_icon(icon, 0, size);
	if (!icon_path) {
		return NULL;
	}
	
	data = find_icon_info(icon_path);
	if (data && data->size == size) {
		data->reference_count++;

		if (GDK_IS_PIXBUF(data->icon.pixbuf)) {
			func(data->icon.pixbuf, NULL,
				data->size, user_data);
		} else if (RSVG_IS_HANDLE(data->icon.svg)) {
			func(NULL, data->icon.svg,
				data->size, user_data);
		}
		g_free(icon_path);
		return data;
	}

	pthread_attr_t thread_attributes;
	pthread_t icon_thread;

	callback_data = g_new(KibaIconFactoryCallbackData, 1);
	callback_data->func = func;
	callback_data->user_data = user_data;

	// check if the surface is in the to make list
	for (l = (GList*) icons_to_make; l != NULL; l = l->next) {
		data = l->data;
		if (strcmp(icon_path, data->uri) == 0 && data->size == size) {
			data->reference_count++;
			// data->callbacks is only used in the main thread, so we can savely access it here ;)
			data->callbacks = g_list_append (data->callbacks, callback_data);
			return data;
		}
	}

	data = g_new0(KibaIconFactoryIconInfo, 1);
	data->uri = icon_path;
	data->size = size;
	data->icon.pixbuf = NULL;
	data->icon.svg = NULL;
	data->callbacks = g_list_append (NULL, callback_data);
	data->reference_count = 1;

	/// MUTEX LOCKED
	pthread_mutex_lock (&icons_mutex); {
		icons_to_make = g_list_append ((GList*) icons_to_make, data);
	} pthread_mutex_unlock (&icons_mutex);
	/// MUTEX UNLOCKED

	if (!icon_load_timeout_id) {
		icon_load_timeout_id = g_timeout_add (50,
			(GSourceFunc) load_icon, NULL);
	}

	if (icon_thread_is_running) {
		return data;
	}

	// We create the thread in the detached state, as we don't need/want
	// to join with it at any point.
	pthread_attr_init (&thread_attributes);
	pthread_attr_setdetachstate (&thread_attributes,
				     PTHREAD_CREATE_DETACHED);
#ifdef _POSIX_THREAD_ATTR_STACKSIZE
	pthread_attr_setstacksize (&thread_attributes, 128*1024);
#endif

	// We set a flag to indicate the thread is running, so we don't create
	// a new one. We don't need to lock a mutex here, as the surface
	// thread isn't running yet. And we know we won't create the thread
	// twice, as we also check thread_starter_id before
	// scheduling this idle function.
	icon_thread_is_running = TRUE;
	pthread_create (&icon_thread, &thread_attributes, thread_start, NULL);

	return data;
}

void
kiba_icon_factory_destroy()
{
	KibaIconFactoryIconInfo *data;
	GList *l;

	if (icon_thread_is_running) {
		icon_thread_is_running = FALSE;
		pthread_exit (NULL);
	}

	for (l = (GList*) icons_to_make; l != NULL; l = (GList*) icons_to_make) {
		icons_to_make = g_list_remove((GList*) icons_to_make, l->data);
		data = l->data;
		destroy_icon_info(data);
	}

	for (l = icons; l != NULL; l = icons) {
		icons = g_list_remove(icons, l->data);
		data = l->data;
		data->reference_count = 1;
		kiba_icon_factory_unload_icon(data);
	}
	g_list_free(icons);
}
