/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <math.h>

#include "kiba.h"
#include "kiba-win.h"
#include "kiba-icon.h"
#include "kiba-utils.h"
#include "kiba-types.h"
#include "kiba-macros.h"

#define FALLBACK_ICON "kiba-dock"

G_DEFINE_TYPE (KibaIcon, kiba_icon, KIBA_TYPE_DRAWABLE_OBJECT)

// property ids
enum {
	_PROP_ROTATE_ANGLE = 1
};

/// private methods

static void
__transform_context(KibaIcon *icon,
		cairo_t *cr)
{
	KibaSize size;
	KibaObject *object = KIBA_OBJECT(icon);
	gfloat dummy_y = object->size.height*object->cfg_scale_y;
	gfloat dummy_x = object->size.width*object->cfg_scale_x;
	
	if (icon->rotate_angle % 180 != 0) {
		size.width = (gfloat) dummy_y;
		size.height = (gfloat) dummy_x;
		cairo_translate(cr,
			-(size.width-dummy_x)/2.0f,
			-(size.height-dummy_y)/2.0f
		);
	} else {
		size.width = (gfloat) dummy_x;
		size.height = (gfloat) dummy_y;
	}
	
	if (icon->rotate_angle != 0) {
		cairo_translate (cr, size.width/2, size.height/2);
		cairo_rotate (cr, 2.0f*M_PI*((gfloat) icon->rotate_angle/360.0f));
		cairo_translate (cr, -size.width/2, -size.height/2);
	}
}

static void
__render_surface_buffer (
		KibaDrawableObject *drawable_object,
		cairo_t *cr)
{
	KibaIcon *icon = KIBA_ICON(drawable_object);
	KibaObject *object = KIBA_OBJECT(icon);
	gfloat width, height;
	
	if (!drawable_object->get_surface) return;
	
	cairo_surface_t *surface = drawable_object->get_surface(drawable_object);
	if (!surface) return;
	
	if (icon->rotate_angle % 180 != 0) {
		width = icon->surface_height;
		height = icon->surface_width;
	} else {
		width = icon->surface_width;
		height = icon->surface_height;
	}
	
	__transform_context(icon, cr);
	cairo_scale(cr,
		(object->alloc.width/width)/object->widget_scale_x,
		(object->alloc.height/height)/object->widget_scale_y);
	
	cairo_set_source_surface(cr, surface, 0, 0);
	cairo_paint(cr);
	cairo_surface_destroy(surface);
}

static void
__update_surface (KibaDrawableObject *drawable_object)
{
	// FIXME: need a surface for custom pixmap
}

static cairo_surface_t*
__get_surface (KibaDrawableObject *drawable_object)
{
	KibaIcon *icon = KIBA_ICON(drawable_object);
	cairo_surface_t *surface = NULL;
	gint surface_width = 0, surface_height = 0;
	
	kiba_cached_icon_get_surface(
		icon->cached_icon,
		MAX(KIBA_OBJECT(icon)->alloc.width,
		    KIBA_OBJECT(icon)->alloc.height),
		&surface, &surface_width, &surface_height);
	icon->surface_width = surface_width;
	icon->surface_height = surface_height;
		
	return cairo_surface_reference(surface);
}

static void
icon_theme_changed_cb (GtkIconTheme *icon_theme,
		gpointer user_data)
{
	KibaIcon *icon = KIBA_ICON(user_data);
	if (icon->file) {
		kiba_icon_set_file (icon, icon->file);
	}
}

/// public methods

/**
 * kiba_icon_set_pixbuf:
 * @icon: #KibaIcon instance.
 * @pixbuf: #GdkPixbuf to set.
 *
 * Sets a custom pixbuf for the icon.
 * Any icon files set before will no be used afterwards.
 */
void
kiba_icon_set_pixbuf(KibaIcon *icon,
		GdkPixbuf *pixbuf)
{
	g_return_if_fail(KIBA_IS_ICON(icon));
	g_return_if_fail(G_IS_OBJECT(pixbuf));
	
	if (G_IS_OBJECT(icon->custom_pixbuf)) {
		g_object_unref(G_OBJECT(icon->custom_pixbuf));
	}
	icon->custom_pixbuf = g_object_ref(G_OBJECT(pixbuf));
	
	kiba_drawable_object_queue_update_surface(KIBA_DRAWABLE_OBJECT(icon));
}

/**
 * kiba_icon_set_file:
 * @icon: #KibaIcon instance.
 * @file: file URI to set.
 *
 * Sets the icon file/name.
 *
 * Returns: TRUE if setting the file was succesfull, otherwise FALSE is returned.
 */
gboolean
kiba_icon_set_file (KibaIcon *icon,
		const gchar *file)
{
	g_return_val_if_fail (KIBA_IS_ICON (icon), FALSE);
	g_return_val_if_fail (file != NULL, FALSE);
	
	cairo_t *cr;
	
	if (icon->file) {
		g_free(icon->file);
	}
	icon->file = g_strdup(file);
	if (KIBA_IS_CACHED_ICON(icon->cached_icon)) {
		g_object_unref(icon->cached_icon);
	}
	cr = kiba_win_cairo_create (kiba_get_win(kiba_get()));
	icon->cached_icon = kiba_cached_icon_new(
		file,
		cr);
	cairo_destroy(cr);
	
	if (G_IS_OBJECT(icon->custom_pixbuf)) {
		g_object_unref(G_OBJECT(icon->custom_pixbuf));
		icon->custom_pixbuf = NULL;
	}

	kiba_drawable_object_queue_update_surface(
		KIBA_DRAWABLE_OBJECT(icon));
	__get_surface (
		KIBA_DRAWABLE_OBJECT(icon));

	return TRUE;
}

/**
 * kiba_icon_get_file:
 * @icon: #KibaIcon instance.
 *
 * Returns @icon name/file.
 *
 * Returns: icon path or name.
 */
const gchar*
kiba_icon_get_file (KibaIcon  *icon)
{
	g_return_val_if_fail (KIBA_IS_ICON (icon), NULL);
	return icon->file;
}

/// Class Functions

/*
 * gtk_widget_unrealize implementation.
 */
static void
kiba_icon_unrealize (GtkWidget *widget)
{
	KibaIcon *icon = KIBA_ICON(widget);

	(* GTK_WIDGET_CLASS (kiba_icon_parent_class)->unrealize) (widget);
	
	if (icon->file) {
		g_free(icon->file);
	}
	if (KIBA_IS_CACHED_ICON(icon->cached_icon)) {
		g_object_unref(icon->cached_icon);
	}
	if (G_IS_OBJECT(icon->custom_pixbuf)) {
		g_object_unref(G_OBJECT(icon->custom_pixbuf));
		icon->custom_pixbuf = NULL;
	}
}

/*
 * gtk_widget_realize implementation.
 */
static void
kiba_icon_realize (GtkWidget *widget)
{
	(* GTK_WIDGET_CLASS (kiba_icon_parent_class)->realize) (widget);

	if (KIBA_ICON(widget)->file) {
		kiba_drawable_object_queue_update_surface(KIBA_DRAWABLE_OBJECT(widget));
	}
}

/*
 * g_object_set_property implementation.
 */
static void
kiba_icon_set_property (GObject *object,
		guint property_id,
		const GValue *value,
		GParamSpec *pspec)
{
	KibaIcon *self = KIBA_ICON(object);

	switch (property_id) {
		case _PROP_ROTATE_ANGLE: {
			gint buf = g_value_get_int (value);
			g_return_if_fail(buf % 90 == 0); // only n*90 degrees rotation supported
			self->rotate_angle = g_value_get_int (value);
			kiba_drawable_object_queue_update_surface(KIBA_DRAWABLE_OBJECT(self));
			break;
		} default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/*
 * g_object_get_property implementation.
 */
static void
kiba_icon_get_property (GObject *object,
		guint property_id,
		GValue *value,
		GParamSpec *pspec)
{
	KibaIcon *self = KIBA_ICON(object);

	switch (property_id) {
		case _PROP_ROTATE_ANGLE:
			g_value_set_int (value, self->rotate_angle);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

static void
kiba_icon_class_init (KibaIconClass *class)
{
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);

	// GObject method implementations
	gobject_class->set_property = kiba_icon_set_property;
	gobject_class->get_property = kiba_icon_get_property;

	widget_class->realize = kiba_icon_realize;
	widget_class->unrealize = kiba_icon_unrealize;
	
	/**
	 * KibaIcon:rotate-angle:
	 *
	 * Rotation angle of the icon.
	 */
	g_object_class_install_property
		(gobject_class,
		 _PROP_ROTATE_ANGLE,
		 g_param_spec_int ("rotate_angle",
		 "Rotate angle",
		 "Rotate angle for icon",
		 -360, 360, 0,
		 G_PARAM_CONSTRUCT|G_PARAM_READWRITE));
}

static void
kiba_icon_init (KibaIcon *icon)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(icon);
	
	drawable_object->update_surface = __update_surface;
	drawable_object->get_surface = __get_surface;
	drawable_object->render_object = __render_surface_buffer;
	
	icon->surface_width = 0;
	icon->surface_height = 0;
	icon->custom_pixbuf = NULL;
	icon->file = NULL;
	icon->cached_icon = NULL;
	
	g_signal_connect (G_OBJECT (gtk_icon_theme_get_default()), "changed",
			G_CALLBACK (icon_theme_changed_cb), icon);
}

/**
 * kiba_icon_new:
 * @object_id: id of the #KibaObject
 * @scale_x: widget scale factor in x direction.
 * @scale_y: widget scale factor in y direction.
 *
 * Creates and returns a #KibaIcon Instance.
 *
 * Returns: #KibaIcon Instance created.
 */
GtkWidget*
kiba_icon_new (const gchar *object_id,
		gfloat scale_x, gfloat scale_y)
{
	GtkWidget *widget =  g_object_new (KIBA_TYPE_ICON,
		"object_id", object_id,
		"rotate_angle", 0,
		"scale_x", scale_x,
		"scale_y", scale_y, NULL);
	return widget;
}

/**
 * kiba_icon_new_from_file:
 * @object_id: id of the #KibaObject
 * @scale_x: widget scale factor in x direction.
 * @scale_y: widget scale factor in y direction.
 * @file: The file URI to use.
 *
 * Creates and returns a #KibaIcon from @file.
 *
 * Returns: #KibaIcon Instance created.
 */
GtkWidget*
kiba_icon_new_from_file (const gchar *object_id,
		gfloat scale_x, gfloat scale_y,
		const gchar *file)
{
	GtkWidget *icon;

	icon = kiba_icon_new (object_id, scale_x, scale_y);
	kiba_icon_set_file(KIBA_ICON(icon), file);

	return icon;
}
