/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <math.h>
#include <limits.h>

#include "marshallers.h"

#include "kiba.h"
#include "kiba-object.h"
#include "kiba-screen.h"
#include "kiba-win.h"
#include "kiba-dock.h"
#include "kiba-macros.h"
#include "kiba-x11.h"

#ifdef USE_LIBSN
  #include "startup-notification.h"
#endif

#define KIBA_OBJECT_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_OBJECT, KibaObjectPrivate))

#define DESTROY_RATE 30

G_DEFINE_ABSTRACT_TYPE (KibaObject, kiba_object, GTK_TYPE_BIN)

typedef struct {
	gchar *object_id;

	gint destroy_id;
	
	guint workTodo;
	guint greyed_count;
	
	gboolean queue_update_scale;
	gboolean queue_update_offset;
	
	GList *processes;
	
	gchar *title;
} KibaObjectPrivate;

// signal ids
enum {
	SIG_URGENT_TOGGLED,
	SIG_WORKING_TOGGLED,
	SIG_GREYED_TOGGLED,
	SIG_ACTIVE_TOGGLED,
	SIG_DRAGGING_TOGGLED,
	SIG_PRELIGHT_TOGGLED,
	SIG_PROCESS_TOGGLED,
	SIG_PROCESS_UPDATED,
	SIG_SCALE_CHANGED,
	SIG_OFFSET_CHANGED,
	SIG_KIBA_OBJECT_LAST
};
static guint kiba_object_signals[SIG_KIBA_OBJECT_LAST] = { 0 };

// property ids
enum {
	_PROP_OBJECT_ID = 1,
	_PROP_SCALE_X,
	_PROP_SCALE_Y,
};

/// Private Methods

static void
__free_feature_list(KibaObject *object, GList **featues)
{
	GList *l;
	for (l = *featues; l != NULL; l = *featues) {
		KibaVisualFeatureObject *o = l->data;
		*featues = g_list_remove (*featues, l->data);
	}
	g_list_free (*featues);
	*featues = NULL;
}

static guint
__get_progress_id(KibaObject *object)
{
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	GList *l;
	guint id_ret;
	gboolean found = FALSE;
	
	for (id_ret = 0; id_ret < UINT_MAX; id_ret++) {
		for (l = kiba_object_get_processes(object); l != NULL; l = l->next) {
			KibaObjectProgress *opro = l->data;
			if (opro->id == id_ret) {
				found = TRUE;
				break;
			}
		}
		if (found) {
			found = FALSE;
			continue;
		} else {
			break;
		}
	}
	
	return id_ret;
}

static gint
__destroy_timeout (gpointer data)
{
	KibaObject *object = data;
	GtkWidget *dock;
	gfloat old_factor;

	dock = GTK_WIDGET(object)->parent;
	
	old_factor = object->destroy_factor;
	object->destroy_factor -= 0.08f;
	object->destroy_factor = MAX(0.0f, object->destroy_factor);
	if (dock) {
		kiba_dock_queue_reposition(KIBA_DOCK(dock));
	}
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(object));
	kiba_object_update_scale_factor (object);

	if (object->destroy_factor == 0.0f) {
		if (dock) {
			gtk_container_remove (GTK_CONTAINER(dock), GTK_WIDGET(object));
		} else {
			gtk_widget_destroy (GTK_WIDGET(object));
		}
		return FALSE;
	}

	return TRUE;
}

static gint
__scale_timeout (gpointer data)
{
	KibaObject *object = data;
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);

	if (priv->destroy_id) {
		return FALSE;
	}

	object->destroy_factor += 0.08f;
	object->destroy_factor = MIN(1.0f, object->destroy_factor);
	if (GTK_WIDGET(object)->parent)
		kiba_dock_queue_reposition(KIBA_DOCK(GTK_WIDGET(object)->parent));
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(object));
	kiba_object_update_scale_factor (object);

	return (object->destroy_factor != 1.0f);
}

//scaling on creation/destruction

void
kiba_object_destroy_factor_func(KibaObject* object,
	gfloat* x, gfloat* y, gpointer data)
{
	*x = object->destroy_factor;
	*y = object->destroy_factor;
}
static void
__destroy_factor_max_func(KibaObject* object,
	gfloat* x, gfloat* y, gpointer data)
{
	*x = 1.0f;
	*y = 1.0f;
}

/// core intern methods

void
kiba_object_set_destroy_factor(KibaObject *object, gfloat factor)
{
	g_return_if_fail (KIBA_IS_OBJECT(object));
	object->destroy_factor = factor;
}

void kiba_object_allocation_set(KibaObject *object)
{
// FIXME: needed?
		g_signal_emit (object,
			kiba_object_signals[SIG_SCALE_CHANGED],
			0, TRUE, TRUE);
}

gboolean
kiba_object_do_update_scale_factor (KibaObject *object)
{
	KibaObjectPrivate *priv;
	gboolean surface_changed;
	gboolean widget_changed;
	gfloat scale_surface_x = 1.0f,
	    scale_surface_y = 1.0f,
	    scale_widget_x = 1.0f,
	    scale_widget_y = 1.0f;
	GList *l;

	priv = KIBA_OBJECT_GET_PRIVATE(object);
	
	if (!priv->queue_update_scale) {
		return FALSE;
	}
	priv->queue_update_scale = FALSE;

	// add current scale factor of all scale methods
	for (l = object->scale_data; l != NULL; l = l->next) {
		KibaObjectScaleData *data = l->data;
		gfloat func_scale_x = 1.0f;
		gfloat func_scale_y = 1.0f;

		data->scale_func (object, &func_scale_x, &func_scale_y, data->data);
		
		if (data->scale_widget) {
			scale_widget_x *= func_scale_x;
			scale_widget_y *= func_scale_y;
		} else {
			scale_surface_x *= func_scale_x;
			scale_surface_y *= func_scale_y;
		}
	}
	
	surface_changed = (object->surface_scale_x != scale_surface_x ||
			object->surface_scale_y != scale_surface_y);
	widget_changed = (object->scale_x/object->surface_scale_x != scale_widget_x ||
			object->scale_y/object->surface_scale_y != scale_widget_y);
	
	object->scale_x = scale_widget_x*scale_surface_x;
	object->scale_y = scale_widget_y*scale_surface_y;
	object->surface_scale_x = scale_surface_x;
	object->surface_scale_y = scale_surface_y;
	object->widget_scale_x = scale_widget_x;
	object->widget_scale_y = scale_widget_y;

	// factor didnt changed
	if (!surface_changed && !widget_changed &&
	            scale_widget_x*scale_surface_x == object->scale_x &&
	            scale_widget_y*scale_surface_y == object->scale_y)
		return FALSE;
	
	if (widget_changed || surface_changed) {
		g_signal_emit (object,
			kiba_object_signals[SIG_SCALE_CHANGED],
			0, widget_changed, surface_changed);
	}

	return TRUE;
}

gboolean
kiba_object_do_update_offset (KibaObject *object)
{
	KibaObjectPrivate *priv;
	gboolean surface_changed;
	gboolean widget_changed;
	gfloat surface_offset_x = 0.0f;
	gfloat surface_offset_y = 0.0f;
	gfloat widget_offset_x = 0.0f;
	gfloat widget_offset_y = 0.0f;
	GList *l;

	priv = KIBA_OBJECT_GET_PRIVATE(object);
	
	if (!priv->queue_update_offset) {
		return FALSE;
	}
	priv->queue_update_offset = FALSE;

	for (l = object->offset_data; l != NULL; l = l->next) {
		KibaObjectOffsetData *data = l->data;
		gfloat func_offset_x = 0.0f;
		gfloat func_offset_y = 0.0f;

		data->place_func (object, &func_offset_x, &func_offset_y, data->data);
		if (data->translate_widget) {
			widget_offset_x += func_offset_x;
			widget_offset_y += func_offset_y;
		} else {
			surface_offset_x += func_offset_x;
			surface_offset_y += func_offset_y;
		}
	}
	
	surface_changed = (object->surface_offset_x != surface_offset_x ||
			object->surface_offset_y != surface_offset_y);
	widget_changed = (object->widget_offset_x != widget_offset_x ||
			object->widget_offset_y != widget_offset_y);
	
	if (!surface_changed && !widget_changed) {
		return TRUE;
	}
	
	object->surface_offset_x = surface_offset_x;
	object->surface_offset_y = surface_offset_y;
	object->widget_offset_x = widget_offset_x;
	object->widget_offset_y = widget_offset_y;
	
	g_signal_emit (object,
		kiba_object_signals[SIG_OFFSET_CHANGED],
		0, widget_changed, surface_changed);

	return TRUE;
}

/// Public methods

/**
 * kiba_object_get_widget_offset:
 * @object: #KibaObject instance.
 * @offset_x: widget offset in x direction.
 * @offset_y: widget offset in y direction.
 *
 * Get the current widget offset.
 */
void
kiba_object_get_widget_offset(KibaObject *object,
		gfloat *offset_x, gfloat *offset_y)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	*offset_x = object->widget_offset_x;
	*offset_y = object->widget_offset_y;
}


/**
 * kiba_object_get_surface_offset:
 * @object: #KibaObject instance.
 * @offset_x: surface offset in x direction.
 * @offset_y: surface offset in y direction.
 *
 * Get the current surface offset.
 */
void
kiba_object_get_surface_offset(KibaObject *object,
		gfloat *offset_x, gfloat *offset_y)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	*offset_x = object->surface_offset_x;
	*offset_y = object->surface_offset_y;
}

/**
 * kiba_object_get_surface_scale:
 * @object: #KibaObject instance.
 * @surface_scale_x: surface scale in x direction.
 * @surface_scale_y: surface scale in y direction.
 *
 * Get the current surface scale.
 */
void
kiba_object_get_surface_scale(KibaObject *object,
		gfloat *surface_scale_x, gfloat *surface_scale_y)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	*surface_scale_x = object->surface_scale_x;
	*surface_scale_y = object->surface_scale_y;
}

/**
 * kiba_object_get_widget_scale:
 * @object: #KibaObject instance.
 * @widget_scale_x: widget scale in x direction.
 * @widget_scale_y: widget scale in y direction.
 *
 * Get the current widget scale.
 */
void
kiba_object_get_widget_scale(KibaObject *object,
		gfloat *widget_scale_x, gfloat *widget_scale_y)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	*widget_scale_x = object->widget_scale_x;
	*widget_scale_y = object->widget_scale_y;
}

/**
 * kiba_object_get_scale:
 * @object: #KibaObject instance.
 * @scale_x: scale in x direction.
 * @scale_y: scale in y direction.
 *
 * Get the current scale.
 */
void
kiba_object_get_scale(KibaObject *object,
		gfloat *scale_x, gfloat *scale_y)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	*scale_x = object->scale_x;
	*scale_y = object->scale_y;
}

/**
 * kiba_object_get_destroy_factor:
 * @object: #KibaObject instance.
 *
 * Returns: The current destroy factor.
 */
float
kiba_object_get_destroy_factor (KibaObject *object)
{
	g_return_val_if_fail (KIBA_IS_OBJECT(object), 1.0f);
	return object->destroy_factor;
}

/**
 * kiba_object_get_state:
 * @object: #KibaObject instance.
 * @state: The state flag.
 *
 * Checks if state flag is set.
 *
 * Returns: Wether @state is set.
 */
gboolean
kiba_object_get_state (KibaObject *object,
		KibaObjectState state)
{
	g_return_val_if_fail(KIBA_IS_OBJECT(object), FALSE);
	return GET_FLAG(object->state, state);
}

/**
 * kiba_object_set_state:
 * @object: #KibaObject instance.
 * @state: The state flag.
 *
 * Sets state flag.
 */
void
kiba_object_set_state (KibaObject *object,
		KibaObjectState state)
{
	KibaObjectPrivate *priv;

	g_return_if_fail(KIBA_IS_OBJECT(object));

	priv = KIBA_OBJECT_GET_PRIVATE (object);

	switch (state) {
	case KIBA_OBJECT_WORKING:
		priv->workTodo++;
		break;
	case KIBA_OBJECT_GREYED:
		priv->greyed_count++;
		break;
	default:
		break;
	}

	if (GET_FLAG(object->state, state)) {
		return;
	}
	SET_FLAG(object->state, state);

	switch (state) {
		case KIBA_OBJECT_WORKING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_WORKING_TOGGLED],
				0, TRUE);
			break;
		}
		case KIBA_OBJECT_ACTIVE: {
			g_signal_emit (object,
				kiba_object_signals[SIG_ACTIVE_TOGGLED],
				0, TRUE);
			break;
		}
		case KIBA_OBJECT_URGENT: {
			g_signal_emit (object,
				kiba_object_signals[SIG_URGENT_TOGGLED],
				0, TRUE);
			break;
		}
		case KIBA_OBJECT_DRAGGING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_DRAGGING_TOGGLED],
				0, TRUE);
			break;
		}
		case KIBA_OBJECT_PRELIGHTED: {
			g_signal_emit (object,
				kiba_object_signals[SIG_PRELIGHT_TOGGLED],
				0, TRUE, TRUE);
			break;
		}
		case KIBA_OBJECT_GREYED: {
			g_signal_emit (object,
				kiba_object_signals[SIG_GREYED_TOGGLED],
				0, TRUE, TRUE);
			break;
		}
		default:
			g_assert_not_reached();
	}
}

/**
 * kiba_object_unset_state:
 * @object: #KibaObject instance.
 * @state: The state flag.
 *
 * Unsets state flag.
 */
void
kiba_object_unset_state (KibaObject *object,
		KibaObjectState state)
{
	KibaObjectPrivate *priv;

	g_return_if_fail(KIBA_IS_OBJECT(object));

	priv = KIBA_OBJECT_GET_PRIVATE (object);

	switch (state) {
	case KIBA_OBJECT_WORKING:
		priv->workTodo--;
		if (priv->workTodo > 0)
			return;
		break;
	case KIBA_OBJECT_GREYED:
		priv->greyed_count = MAX(0, priv->greyed_count-1);
		if (priv->greyed_count > 0)
			return;
		break;
	default:
		break;
	}

	if (!GET_FLAG(object->state, state)) {
		return;
	}
	DEL_FLAG(object->state, state);

	switch (state) {
		case KIBA_OBJECT_WORKING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_WORKING_TOGGLED],
				0, FALSE);
			break;
		}
		case KIBA_OBJECT_ACTIVE: {
			g_signal_emit (object,
				kiba_object_signals[SIG_ACTIVE_TOGGLED],
				0, FALSE);
			break;
		}
		case KIBA_OBJECT_URGENT: {
			g_signal_emit (object,
				kiba_object_signals[SIG_URGENT_TOGGLED],
				0, FALSE);
			break;
		}
		case KIBA_OBJECT_DRAGGING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_DRAGGING_TOGGLED],
				0, FALSE);
			break;
		}
		case KIBA_OBJECT_PRELIGHTED: {
			g_signal_emit (object,
				kiba_object_signals[SIG_PRELIGHT_TOGGLED],
				0, FALSE, TRUE);
			break;
		}
		case KIBA_OBJECT_GREYED: {
			g_signal_emit (object,
				kiba_object_signals[SIG_GREYED_TOGGLED],
				0, FALSE, TRUE);
			break;
		}
		default:
			g_assert_not_reached();
	}
}

/**
 * kiba_object_set_tooltip_text:
 * @object: #KibaObject instance.
 *
 */
void
kiba_object_set_tooltip_text(KibaObject *object,
		const gchar *text)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
#if GTK_CHECK_VERSION(2,12,0) == TRUE
	gtk_widget_set_tooltip_text(GTK_WIDGET(object), text);
#endif
}

/**
 * kiba_object_launch:
 * @object: #KibaObject Instance.
 * @command: command to launch.
 *
 * The method executes the given command.
 * Additional it passes the command to libsn, to monitor the launch process.
 *
 * Returns: The PID or -1.
 */
int
kiba_object_launch(KibaObject *object,
		const gchar *command)
{
	g_return_val_if_fail (KIBA_IS_OBJECT(object), -1);
	g_return_val_if_fail (command != NULL, -1);
	
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
#ifdef USE_LIBSN
	// monitor this launch with libsn
	kiba_sn_launch(command, G_OBJECT(object));
#endif
	return kiba_launch_programm(command);
}

/**
 * kiba_object_set_icon_geometry:
 * @object: #KibaObject instance.
 * @window: #KibaWindow to set the _NET_WM_ICON property on.
 *
 * Sets the _NET_WM_ICON property on @window, for minimze/restore animations.
 */
void
kiba_object_set_icon_geometry (KibaObject *object,
		KibaWindow *window)
{
	GtkWidget *dock;
	GtkAllocation geometry;
	gint fX, fY;

	g_return_if_fail (KIBA_IS_WINDOW(window));
	g_return_if_fail (KIBA_IS_OBJECT(object));

	dock = GTK_WIDGET(object)->parent;

	fX = (gint) object->alloc.x;
	fY = (gint) object->alloc.y;

	geometry.width = GTK_WIDGET(object)->allocation.width;
	geometry.height = GTK_WIDGET(object)->allocation.height;
	geometry.x = dock->parent->allocation.x + fX;
	geometry.y = dock->parent->allocation.y + fY;

	kiba_window_set_minimize_restore_point(window,
			geometry.x, geometry.y,
			geometry.width, geometry.height);
}

/**
 * kiba_object_show_title:
 * @object: #KibaObject instance.
 *
 * Shows the object title.
 */
void
kiba_object_show_title (KibaObject *object)
{
	KibaTitle *title;
	KibaDock *dock;

	g_return_if_fail (KIBA_IS_OBJECT(object));

	title = kiba_get_title(kiba_get());
	dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	if (!kiba_title_get_flag(title, KIBA_TITLE_DRAWABLE)) {
		return;
	}

	kiba_title_show_object (title,
		GTK_WIDGET(object), dock->d.position);
}

/**
 * kiba_object_hide_title:
 * @object: #KibaObject instance.
 *
 * Hides the object title.
 */
void
kiba_object_hide_title (KibaObject *object)
{
	g_return_if_fail (KIBA_IS_OBJECT(object));
	kiba_title_hide_object (kiba_get_title(kiba_get()), GTK_WIDGET(object));
}

/**
 * kiba_object_set_title:
 * @object: #KibaObject instance.
 * @text: object title.
 *
 * Sets the object title text.
 */
void
kiba_object_set_title (KibaObject *object,
		const gchar *text)
{
	KibaObjectPrivate *priv;

	g_return_if_fail (KIBA_IS_OBJECT(object));

	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (priv->title) {
		g_free(priv->title);
	}
	priv->title = g_strdup(text);
	
	if (GTK_WIDGET(object)->parent) {
		kiba_title_set_object_text (kiba_get_title(kiba_get()),
			GTK_WIDGET(object), text);
	}
}

/**
 * kiba_object_get_title:
 * @object: #KibaObject instance.
 *
 * Returns: object title.
 */
const gchar*
kiba_object_get_title (KibaObject *object)
{
	g_return_val_if_fail (KIBA_IS_OBJECT (object), NULL);
	return kiba_title_get_object_text(kiba_get_title(kiba_get()), GTK_WIDGET(object));
}

/**
 * kiba_object_get_destroying:
 * @object: #KibaObject instance.
 *
 * Returns: Wether the object is about to be destroyed.
 */
gboolean
kiba_object_get_destroying (KibaObject *object)
{
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	return KIBA_OBJECT_GET_PRIVATE (object)->destroy_id != 0;
}

/**
 * kiba_object_destroy:
 * @object: #KibaObject instance.
 *
 * Destroys the #KibaObject, scaling the object down and unrealizing it.
 */
void
kiba_object_destroy (KibaObject *object)
{
	KibaObjectPrivate *priv;

	g_return_if_fail (KIBA_IS_OBJECT(object));

	priv = KIBA_OBJECT_GET_PRIVATE (object);

	if (!priv->destroy_id)
		priv->destroy_id = g_timeout_add (DESTROY_RATE, __destroy_timeout, object);

	if (kiba_object_get_state(object, KIBA_OBJECT_ACTIVE))
		kiba_object_unset_state(object, KIBA_OBJECT_ACTIVE);
}

/**
 * kiba_object_make_sensitive:
 * @object: #KibaObject instance.
 * @sensitive: should the object be sensitive?
 * 
 * Sets the sensitive flag on @object.
 * Only sensitive objects can receive events.
 */
void
kiba_object_make_sensitive(KibaObject *object,
		gboolean sensitive)
{
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	GtkWidget *widget = GTK_WIDGET(object);
	
	if (sensitive) {
		if (!GTK_WIDGET_SENSITIVE(object)) {
			g_object_set(G_OBJECT(object), "sensitive", TRUE, NULL);
			kiba_object_unset_state(object, KIBA_OBJECT_GREYED);
		}
	} else {
		if (!GET_FLAG(object->state, KIBA_OBJECT_DRAGGING) && GTK_WIDGET_SENSITIVE(object)) {
			g_object_set(G_OBJECT(object), "sensitive", FALSE, NULL);
			kiba_object_set_state(object, KIBA_OBJECT_GREYED);
		}
	}
}

/**
 * kiba_object_remove_offset_function_by_data:
 * @object: #KibaObject instance.
 * @user_data: custom user data.
 *
 * Removes offset function.
 */
void
kiba_object_remove_offset_function_by_data (KibaObject *object,
		gpointer user_data)
{
	KibaObjectOffsetData *data;
	GList *l;

	g_return_if_fail (KIBA_IS_OBJECT(object));

	for (l = object->offset_data; l != NULL; l = l->next) {
		data = l->data;
		if (data->data == user_data) {
			g_free (data);
			object->offset_data = g_list_remove (object->offset_data, l->data);
			gtk_widget_queue_resize_no_redraw(GTK_WIDGET(object));
			return;
		}
	}
}

/**
 * kiba_object_add_offset_function:
 * @object: #KibaObject instance.
 * @place_func: surface offset method.
 * @max_func: method to get the maximal offset.
 * @translate_widget: should the widget be translated?
 * @user_data: custom user data.
 *
 * Adds offset function, for object positioning.
 */
void
kiba_object_add_offset_function (KibaObject *object,
		KibaObjectOffsetFunc place_func,
		KibaObjectMaxOffsetFunc max_func,
		gboolean translate_widget,
		gpointer user_data)
{
	KibaObjectOffsetData *data;
	GList *l;

	g_return_if_fail (KIBA_IS_OBJECT(object));

	for (l = object->offset_data; l != NULL; l = l->next) {
		data = l->data;
		if (data->place_func == place_func) {
			data->data = user_data;
			return;
		}
	}

	data = g_new0 (KibaObjectOffsetData, 1);
	data->place_func = place_func;
	data->max_func = max_func;
	data->translate_widget = translate_widget;
	data->data = user_data;

	object->offset_data = g_list_append (object->offset_data, data);
	
	kiba_object_update_offset(object);
}

/**
 * kiba_object_remove_scale_function_by_data:
 * @object: #KibaObject instance.
 * @user_data: custom user data.
 *
 * Removes scale function.
 */
void
kiba_object_remove_scale_function_by_data (
		KibaObject *object,
		gpointer    user_data)
{
	KibaObjectScaleData *data;
	GList *l;

	g_return_if_fail (KIBA_IS_OBJECT(object));

	for (l = object->scale_data; l != NULL; l = l->next) {
		data = l->data;
		if (data->data == user_data) {
			g_free (data);
			object->scale_data = g_list_remove (object->scale_data, l->data);
			gtk_widget_queue_resize_no_redraw(GTK_WIDGET(object));
			kiba_object_remove_scale_function_by_data (object, user_data);
			return;
		}
	}
	kiba_object_update_scale_factor(object);
}

/**
 * kiba_object_add_scale_function:
 * @object: #KibaObject instance.
 * @scale_func: scale method.
 * @max_func: method to return the maximal scale factor.
 * @scale_widget: TRUE to scale #KibaObject widget, otherwise FALSE to scale only the surface.
 * @user_data: custom user data.
 *
 * Adds scale function, for scale animations.
 */
void
kiba_object_add_scale_function (KibaObject *object,
		KibaObjectScaleFunc scale_func,
		KibaObjectMaxScaleFunc max_func,
		gboolean scale_widget,
		gpointer user_data)
{
	KibaObjectScaleData *data;
	GList *l;

	g_return_if_fail (KIBA_IS_OBJECT(object));

	for (l = object->scale_data; l != NULL; l = l->next) {
		data = l->data;
		if (data->scale_func == scale_func) {
			data->data = user_data;
			data->scale_widget = scale_widget;
			return;
		}
	}

	data = g_new0 (KibaObjectScaleData, 1);
	data->scale_func = scale_func;
	data->max_func = max_func;
	data->scale_widget = scale_widget;
	data->data = user_data;

	object->scale_data = g_list_append (object->scale_data, data);
	
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(object));
	kiba_object_update_scale_factor(object);
}

/**
 * kiba_object_is_on_dock:
 * @object: #KibaObject instance.
 *
 * Returns whether object is on the dock.
 *
 * Returns: TRUE if @object is on the dock, otherwise FALSE is returned.
 **/
gboolean
kiba_object_is_on_dock (KibaObject *object)
{
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	return object->position == KIBA_OBJECT_ON_DOCK;
}

/**
 * kiba_object_update_scale_factor:
 * @object: #KibaObject instance.
 *
 * Queues scale factor recalculation.
 */
void
kiba_object_update_scale_factor (KibaObject *object)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	
	KibaObjectPrivate *priv =
		KIBA_OBJECT_GET_PRIVATE(object);
	priv->queue_update_scale = TRUE;
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(object));
}

/**
 * kiba_object_update_offset:
 * @object: #KibaObject instance.
 *
 * Queues offset recalculation.
 */
void
kiba_object_update_offset (KibaObject *object)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	
	KibaObjectPrivate *priv =
		KIBA_OBJECT_GET_PRIVATE(object);
	priv->queue_update_offset = TRUE;
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(object));
}

/**
 * kiba_object_disable_feature:
 * @object: #KibaObject instance.
 * @@feature: #KibaVisualFeatureObject to enable.
 * 
 * Enables @feature.
 */
void
kiba_object_enable_feature(KibaObject *object,
		KibaVisualFeatureObject *feature_object)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	g_return_if_fail(feature_object != NULL);
	
	switch(kiba_visual_feature_type(feature_object->feature)) {
	case KIBA_VISUAL_FEATURE_MORPH:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_morph_front = g_list_prepend(
				object->features_morph_front, feature_object);
		} else {
			object->features_morph_back = g_list_prepend(
				object->features_morph_back, feature_object);
		}
		break;
	case KIBA_VISUAL_FEATURE_ONTOP:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_ontop_front = g_list_prepend(
				object->features_ontop_front, feature_object);
		} else {
			object->features_ontop_back = g_list_prepend(
				object->features_ontop_back, feature_object);
		}
		break;
	case KIBA_VISUAL_FEATURE_ONTOP_OF_DOCK:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_ontop_dock_front = g_list_prepend(
				object->features_ontop_dock_front, feature_object);
		} else {
			object->features_ontop_dock_back = g_list_prepend(
				object->features_ontop_dock_back, feature_object);
		}
		break;
	case KIBA_VISUAL_FEATURE_GIMMICK:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_gimmick_front = g_list_prepend(
				object->features_gimmick_front, feature_object);
		} else {
			object->features_gimmick_back = g_list_prepend(
				object->features_gimmick_back, feature_object);
		}
		break;
	default:
		g_assert_not_reached();
	}
}

/**
 * kiba_object_disable_feature:
 * @object: #KibaObject instance.
 * @feature_object: #KibaVisualFeatureObject to disble.
 * 
 * Disables @feature_object.
 */
void
kiba_object_disable_feature(KibaObject *object,
		KibaVisualFeatureObject *feature_object)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	g_return_if_fail(feature_object != NULL);
	
	switch(kiba_visual_feature_type(feature_object->feature)) {
	case KIBA_VISUAL_FEATURE_MORPH:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_morph_front = g_list_remove(
				object->features_morph_front, feature_object);
		} else {
			object->features_morph_back = g_list_remove(
				object->features_morph_back, feature_object);
		}
		break;
	case KIBA_VISUAL_FEATURE_ONTOP:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_ontop_front = g_list_remove(
				object->features_ontop_front, feature_object);
		} else {
			object->features_ontop_back = g_list_remove(
				object->features_ontop_back, feature_object);
		}
		break;
	case KIBA_VISUAL_FEATURE_ONTOP_OF_DOCK:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_ontop_dock_front = g_list_remove(
				object->features_ontop_dock_front, feature_object);
		} else {
			object->features_ontop_dock_back = g_list_remove(
				object->features_ontop_dock_back, feature_object);
		}
		break;
	case KIBA_VISUAL_FEATURE_GIMMICK:
		if (kiba_visual_feature_get_front(feature_object->feature)) {
			object->features_gimmick_front = g_list_remove(
				object->features_gimmick_front, feature_object);
		} else {
			object->features_gimmick_back = g_list_remove(
				object->features_gimmick_back, feature_object);
		}
		break;
	default:
		g_assert_not_reached();
	}
}

/**
 * kiba_object_get_processes:
 * @object: #KibaObject instance.
 *
 * Returns: list of active progresses on the object.
 **/
GList*
kiba_object_get_processes(KibaObject *object)
{
	g_return_val_if_fail(KIBA_IS_OBJECT(object), NULL);
	return KIBA_OBJECT_GET_PRIVATE (object)->processes;
}

/**
 * kiba_object_start_progress:
 * @object: #KibaObject instance.
 *
 * Creates and starts a progress on @object.
 * 
 * Returns: #KibaObjectProgress instance created.
 **/
KibaObjectProgress*
kiba_object_start_progress(KibaObject *object)
{
	KibaObjectProgress *progress;
	
	g_return_val_if_fail(KIBA_IS_OBJECT(object), NULL);
	
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	progress = g_new0(KibaObjectProgress, 1);
	progress->id = __get_progress_id(object);
	progress->value = 0;
	
	priv->processes = g_list_prepend(priv->processes, progress);
	
	g_signal_emit (object,
		kiba_object_signals[SIG_PROCESS_TOGGLED],
		0, progress->id, TRUE);
	
	return progress;
}

/**
 * kiba_object_stop_progress:
 * @object: #KibaObject instance.
 * @progress: #KibaObjectProgress to stop.
 *
 * Stops a previously started progress.
 **/
void
kiba_object_stop_progress(KibaObject *object,
		KibaObjectProgress *progress)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	g_return_if_fail(progress != NULL);
	
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	g_signal_emit (object,
		kiba_object_signals[SIG_PROCESS_TOGGLED],
		0, progress->id, FALSE);
	
	priv->processes = g_list_remove(priv->processes, progress);
	g_free(progress);
}

/**
 * kiba_object_update_progress:
 * @object: #KibaObject instance.
 * @progress: #KibaObjectProgress to update.
 * @value: new value of the process.
 * 
 * Sets @value on @progress.
 **/ 
void
kiba_object_update_progress(KibaObject *object,
		KibaObjectProgress *progress, gint value)
{
	g_return_if_fail(KIBA_IS_OBJECT(object));
	g_return_if_fail(progress != NULL);
	g_return_if_fail(value >= 0 && value <= 100);
	
	progress->value = value;
	g_signal_emit (object,
		kiba_object_signals[SIG_PROCESS_UPDATED],
		0, progress->id, value);
}

/// Class Implementations

/*
 * gtk_widget_size_request implementation.
 */
static void
kiba_object_size_request (GtkWidget *widget,
		GtkRequisition *requisition)
{
	KibaObject *object = KIBA_OBJECT(widget);
	// widget scale factors
	gfloat scale_x;
	gfloat scale_y;
	
	scale_x = object->widget_scale_x*object->cfg_scale_x;
	scale_y = object->widget_scale_y*object->cfg_scale_y;

	requisition->width = (gint) (object->size.width*scale_x);
	requisition->height = (gint) (object->size.height*scale_y);
}


/*
 * gtk_widget_unrealize implementation.
 */
static void
kiba_object_unrealize (GtkWidget *widget)
{
	KibaObject *object = KIBA_OBJECT(widget);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (widget);
	GList *l;

	(* GTK_WIDGET_CLASS (kiba_object_parent_class)->unrealize) (widget);

	if (priv->destroy_id)
		g_source_remove (priv->destroy_id);
	priv->destroy_id = 0;

	for (l = object->scale_data; l != NULL; l = object->scale_data) {
		KibaObjectScaleData *data = l->data;
		g_free (data);
		object->scale_data = g_list_remove (object->scale_data, l->data);
	}
	g_list_free (object->scale_data);
	object->scale_data = NULL;

	for (l = object->offset_data; l != NULL; l = object->offset_data) {
		KibaObjectOffsetData *data = l->data;
		g_free (data);
		object->offset_data = g_list_remove (object->offset_data, l->data);
	}
	g_list_free (object->offset_data);
	object->offset_data = NULL;

	for (l = priv->processes; l != NULL; l = priv->processes) {
		KibaObjectProgress *data = l->data;
		g_free (data);
		priv->processes = g_list_remove (priv->processes, l->data);
	}
	g_list_free (priv->processes);
	priv->processes = NULL;
	
	if (priv->title) {
		g_free(priv->title);
		priv->title = NULL;
	}

	__free_feature_list(object, &(object->features_morph_front));
	__free_feature_list(object, &(object->features_morph_back));
	__free_feature_list(object, &(object->features_ontop_front));
	__free_feature_list(object, &(object->features_ontop_back));
	__free_feature_list(object, &(object->features_ontop_dock_front));
	__free_feature_list(object, &(object->features_ontop_dock_back));
	__free_feature_list(object, &(object->features_gimmick_front));
	__free_feature_list(object, &(object->features_gimmick_back));

	kiba_title_remove_object (kiba_get_title(kiba_get()), GTK_WIDGET(object));
}

/*
 * gtk_widget_realize implementation.
 */
static void
kiba_object_realize (GtkWidget *widget)
{
	// this method must be implemented by the subclass
	g_warning ("KibaObjectClass::realize not implemented for `%s'",
		g_type_name (G_TYPE_FROM_INSTANCE (widget)));
}

/*
 * gtk_widget_parent_set implementation.
 */
static void
kiba_object_parent_set (GtkWidget *widget,
		GtkWidget *old_parent)
{
	KibaObject *object = KIBA_OBJECT(widget);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	if (widget->parent) {
		KibaDock *dock = KIBA_DOCK(widget->parent);
		if (kiba_object_get_destroy_factor(object) == 0.0f)
			g_timeout_add (DESTROY_RATE, __scale_timeout, object);
		kiba_title_add_object (kiba_get_title(kiba_get()),
			widget, dock->d.position);
		if (priv->title) {
			kiba_title_set_object_text (kiba_get_title(kiba_get()),
				widget, priv->title);
		}
	}
}

/*
 * g_object_set_property implementation.
 */
static void
kiba_object_set_property (GObject *object,
		guint property_id,
		const GValue *value,
		GParamSpec *pspec)
{
	KibaObject *self = KIBA_OBJECT(object);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (self);

	switch (property_id) {
		case _PROP_OBJECT_ID:
			priv->object_id = g_value_dup_string (value);
			break;
		case _PROP_SCALE_X:
			self->cfg_scale_x = g_value_get_float(value);
			gtk_widget_queue_resize_no_redraw(GTK_WIDGET(self));
			break;
		case _PROP_SCALE_Y:
			self->cfg_scale_y = g_value_get_float(value);
			gtk_widget_queue_resize_no_redraw(GTK_WIDGET(self));
			g_signal_emit (self,
				kiba_object_signals[SIG_SCALE_CHANGED],
				0, TRUE, FALSE);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/*
 * g_object_get_property implementation.
 */
static void
kiba_object_get_property (GObject *object,
		guint property_id,
		GValue *value,
		GParamSpec *pspec)
{
	KibaObject *self = KIBA_OBJECT(object);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (self);

	switch (property_id) {
		case _PROP_OBJECT_ID:
			g_value_set_string (value, priv->object_id);
			break;
		case _PROP_SCALE_X:
			g_value_set_float (value, self->cfg_scale_x);
			break;
		case _PROP_SCALE_Y:
			g_value_set_float (value, self->cfg_scale_y);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/*** Class init ***/

static void
kiba_object_class_init (KibaObjectClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

	g_type_class_add_private (class, sizeof (KibaObjectPrivate));

	// GObject method implementations
	gobject_class->set_property = kiba_object_set_property;
	gobject_class->get_property = kiba_object_get_property;

	// GtkWidget method implementations
	widget_class->realize = kiba_object_realize;
	widget_class->unrealize = kiba_object_unrealize;
	widget_class->size_request = kiba_object_size_request;
	widget_class->parent_set = kiba_object_parent_set;

	// properties
	/**
	 * KibaObject:object-id:
	 *
	 * Object id, must be unique for a dock.
	 */
	g_object_class_install_property
		(gobject_class,
		 _PROP_OBJECT_ID,
		 g_param_spec_string ("object_id",
		 "Object id",
		 "Object id, must be unique for a dock.",
		 NULL,
		 G_PARAM_CONSTRUCT|G_PARAM_READWRITE));
	/**
	 * KibaObject:scale-x:
	 *
	 * Scale factor in x direction.
	 */
	g_object_class_install_property
		(gobject_class,
		 _PROP_SCALE_X,
		 g_param_spec_float ("scale_x",
		 "Scale factor in x direction",
		 "Scale factor in x direction",
		 0.0f, 4.0f, 1.0f,
		 G_PARAM_CONSTRUCT|G_PARAM_READWRITE));
	/**
	 * KibaObject:scale-y:
	 *
	 * Scale factor in y direction.
	 */
	g_object_class_install_property
		(gobject_class,
		 _PROP_SCALE_Y,
		 g_param_spec_float ("scale_y",
		 "Scale factor in y direction",
		 "Scale factor in y direction",
		 0.0f, 4.0f, 1.0f,
		 G_PARAM_CONSTRUCT|G_PARAM_READWRITE));

	// signals
	/**
	 * KibaObject::urgent-toggled:
	 * @object: the object on which the signal is emitted
	 * @urgent: toggled on?
	 *
	 * The urgent-changed signal is emitted when the urgent state changed
	 * for @object.
	 */
	kiba_object_signals[SIG_URGENT_TOGGLED] =
		g_signal_new ("urgent-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, urgent_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	/**
	 * KibaObject::active-toggled:
	 * @object: the object on which the signal is emitted
	 * @active: toggled on?
	 *
	 * The active-changed signal is emitted when the active state changed
	 * for @object.
	 */
	kiba_object_signals[SIG_ACTIVE_TOGGLED] =
		g_signal_new ("active-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, active_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	/**
	 * KibaObject::working-toggled:
	 * @object: the object on which the signal is emitted
	 * @working: toggled on?
	 *
	 * The working-changed signal is emitted when the working state changed
	 * for @object.
	 */
	kiba_object_signals[SIG_WORKING_TOGGLED] =
		g_signal_new ("working-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, working_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	/**
	 * KibaObject::process-toggled:
	 * @object: the object on which the signal is emitted
	 * @progress_id: progress identification number.
	 * @toggled: toggled on?
	 *
	 * The process-changed signal is emitted when the process state changed
	 * for @object.
	 */
	kiba_object_signals[SIG_PROCESS_TOGGLED] =
		g_signal_new ("process-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, process_toggled),
			NULL, NULL,
			_kiba_marshal_VOID__INT_BOOLEAN,
			G_TYPE_NONE, 2,
			G_TYPE_INT,
			G_TYPE_BOOLEAN);
	/**
	 * KibaObject::process-updated:
	 * @object: the object on which the signal is emitted
	 * @progress_id: progress identification number.
	 * @value: new progress value.
	 *
	 * The process-updated signal is emitted when a process changed
	 * for @object.
	 */
	kiba_object_signals[SIG_PROCESS_UPDATED] =
		g_signal_new ("process-updated",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, process_updated),
			NULL, NULL,
			_kiba_marshal_VOID__INT_INT,
			G_TYPE_NONE, 2,
			G_TYPE_INT,
			G_TYPE_INT);
	/**
	 * KibaObject::dragging-toggled:
	 * @object: the object on which the signal is emitted
	 * @dragging: toggled on?
	 *
	 * The dragging-changed signal is emitted when the dragging state changed
	 * for @object.
	 */
	kiba_object_signals[SIG_DRAGGING_TOGGLED] =
		g_signal_new ("dragging-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, dragging_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	/**
	 * KibaObject::prelight-toggled:
	 * @object: the object on which the signal is emitted
	 * @prelight: toggled on?
	 *
	 * The prelight-changed signal is emitted when the prelight state changed
	 * for @object.
	 */
	kiba_object_signals[SIG_PRELIGHT_TOGGLED] =
		g_signal_new ("prelight-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, prelight_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	/**
	 * KibaObject::greyed-toggled:
	 * @object: the object on which the signal is emitted
	 * @greyed: toggled on?
	 *
	 * The greyed-changed signal is emitted when the greyed state changed
	 * for @object.
	 */
	kiba_object_signals[SIG_GREYED_TOGGLED] =
		g_signal_new ("greyed-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, prelight_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	/**
	 * KibaObject::scale-changed:
	 * @object: the object on which the signal is emitted
	 * @surface_changed: surface scale changed?
	 * @widget_offset: widget scale changed?
	 *
	 * The scale-changed signal is emitted when a new scale was calculated
	 * for @object.
	 */
	kiba_object_signals[SIG_SCALE_CHANGED] =
		g_signal_new ("scale-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, scale_changed),
			NULL, NULL,
			_kiba_marshal_VOID__BOOLEAN_BOOLEAN,
			G_TYPE_NONE, 2,
			G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
	/**
	 * KibaObject::offset-changed:
	 * @object: the object on which the signal is emitted
	 * @surface_changed: surface offset changed?
	 * @widget_offset: widget offset changed?
	 *
	 * The offset-changed signal is emitted when a new offset was calculated
	 * for @object.
	 */
	kiba_object_signals[SIG_OFFSET_CHANGED] =
		g_signal_new ("offset-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, offset_changed),
			NULL, NULL,
			_kiba_marshal_VOID__BOOLEAN_BOOLEAN,
			G_TYPE_NONE, 2,
			G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
}

static void
kiba_object_init (KibaObject *object)
{
	static const GtkTargetEntry targets[] = {
		{ "text/plain", 0, 0 }
	};

	KibaObjectPrivate *priv;

	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (object), GTK_NO_WINDOW);

	priv = KIBA_OBJECT_GET_PRIVATE (object);

	gtk_widget_add_events (GTK_WIDGET (object),
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	gtk_widget_set_redraw_on_allocate (
		GTK_WIDGET(object), FALSE);

	object->position = KIBA_OBJECT_LEFT_OF_DOCK;
	object->cfg_scale_x = 1.0f;
	object->cfg_scale_y = 1.0f;
	object->alloc = (KibaAllocation) {0.0f, 0.0f, 0.0f, 0.0f};
	object->size = (GtkRequisition) {28, 28};
	object->scale_data = NULL;
	object->offset_data = NULL;
	object->scale_x = 1.0f;
	object->scale_y = 1.0f;
	object->surface_scale_x = 1.0f;
	object->surface_scale_y = 1.0f;
	object->widget_scale_x = 1.0f;
	object->widget_scale_y = 1.0f;
	object->surface_offset_x = 0.0f;
	object->surface_offset_y = 0.0f;
	object->widget_offset_x = 0.0f;
	object->widget_offset_y = 0.0f;
	object->destroy_factor = 0.0f;
	object->features_morph_front = NULL;
	object->features_morph_back = NULL;
	object->features_ontop_front = NULL;
	object->features_ontop_back = NULL;
	object->features_ontop_dock_front = NULL;
	object->features_ontop_dock_back = NULL;
	object->features_gimmick_front = NULL;
	object->features_gimmick_back = NULL;

	priv->workTodo = 0;
	priv->greyed_count = 0;
	priv->title = NULL;
	priv->queue_update_scale = TRUE;
	priv->queue_update_offset = TRUE;
	
	// enable drag and drop
	gtk_drag_dest_set (GTK_WIDGET (object),
			GTK_DEST_DEFAULT_ALL,
			targets, G_N_ELEMENTS (targets),
			GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);
	gtk_drag_dest_add_uri_targets (GTK_WIDGET (object));
	
	kiba_object_add_scale_function (object,
		kiba_object_destroy_factor_func,
                __destroy_factor_max_func,
                TRUE, NULL);
}
