/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"
#include <libintl.h>
#include <kiba-icon.h>

#include "kiba-separator.h"
#include "kiba-dock.h"

#define SEPARATOR_ICON  KIBA_ICON_PATH"/separator.svg"

#define _(string)  gettext (string)

struct _KibaSeparator {
	KibaIcon parent;
};

struct _KibaSeparatorClass {
	KibaIconClass parent_class;
};

G_DEFINE_TYPE (KibaSeparator, kiba_separator, KIBA_TYPE_ICON)

static void
__update_size(KibaSeparator *separator,
		KibaDock *dock)
{
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	KibaIcon *icon = KIBA_ICON(separator);
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		g_object_set (G_OBJECT (separator),
			"rotate_angle", -(d->halignment-1)*90,
			"scale_x", 1.0f, 
			"scale_y", (float) icon->surface_width/icon->surface_height, 
			NULL);
		kiba_drawable_object_queue_update_surface(KIBA_DRAWABLE_OBJECT(separator));
	} else {
		g_object_set (G_OBJECT (separator),
			"rotate_angle", ((d->valignment-2)/2)*180,
			"scale_x", (float) icon->surface_width/icon->surface_height, 
			"scale_y", 1.0f, 
			NULL);
		kiba_drawable_object_queue_update_surface(KIBA_DRAWABLE_OBJECT(separator));
	}
}

static void
__position_changed_cb(KibaDock *dock,
		KibaPosition position,
		gpointer data)
{
	__update_size(KIBA_SEPARATOR(data), dock);
}

static gboolean
__remove_separator_cb (GtkWidget  *widget,
				     gpointer    data) {
	kiba_object_destroy (KIBA_OBJECT(data));
	return TRUE;
}

/*** public methods ***/

/*** custom implementations ***/

static void
kiba_separator_parent_set(GtkWidget *widget, GtkWidget *old_parent)
{
	if (old_parent) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(old_parent),
			G_CALLBACK(__position_changed_cb), widget);
	}
	if (widget->parent) {
		KibaDock *dock = KIBA_DOCK(widget->parent);
		g_signal_connect (G_OBJECT(dock), "position-changed",
			G_CALLBACK(__position_changed_cb), widget);
		__position_changed_cb(dock, dock->d.position, widget);
	}
	
	(* GTK_WIDGET_CLASS (kiba_separator_parent_class)->parent_set) (widget, old_parent);
}

static gboolean
kiba_separator_button_press_event (GtkWidget *widget,
		GdkEventButton *event)
{
	GtkWidget *menu;
	GtkWidget *image, *item;
	
	if (event->button != 3)
		return TRUE;
	
	menu = gtk_menu_new ();
	
	item  = gtk_image_menu_item_new_with_label (_("Remove Separator"));
	image = gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
	gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), item);
	g_signal_connect (G_OBJECT (item), "activate",
		G_CALLBACK (__remove_separator_cb), widget);
	gtk_widget_show (item);
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

/// class init/fini

static void
kiba_separator_class_init (KibaSeparatorClass *class)
{
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	widget_class->parent_set = kiba_separator_parent_set;
	widget_class->button_press_event = kiba_separator_button_press_event;
}

static void
kiba_separator_init (KibaSeparator *separator)
{
	GtkWidget *widget = GTK_WIDGET(separator);
}

/**
 * kiba_separator_new:
 * @object_id: id of the #KibaObject
 *
 * Creates and returns a #KibaSeparator Instance.
 *
 * Returns: #KibaSeparator object created.
 */
GtkWidget* kiba_separator_new (const gchar *object_id)
{
	GtkWidget *widget =  g_object_new (KIBA_TYPE_SEPARATOR,
		"object_id", object_id,
		"rotate_angle", 0,
		"scale_x", 1.0f,
		"scale_y", 1.0f,
		NULL);
	kiba_icon_set_file(KIBA_ICON(widget), SEPARATOR_ICON);
	return widget;
}
