/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <math.h>
#include <string.h>
#include <glib.h>

#include "kiba-title.h"
#include "kiba-screen.h"
#include "kiba-cairo-utils.h"
#include "kiba-debug.h"
#include "kiba-macros.h"
#include "kiba-prefs.h"
#include "kiba-transparent-window.h"

#define BACKGROUND_STEPSIZE  0.02f
#define FADE_FPS 100
#define TITLE_OBJECT_DISTANCE 9
#define MAX_FPS 100

#define WINDOW_ROLE "kiba-title"
#define WINDOW_ICON_NAME "kiba-dock"

#define KIBA_TITLE_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_TITLE, KibaTitlePrivate))

// preference keys
#define PREFERENCES_ENABLE_KEY "title_enable"
#define PREFERENCES_HIGHLIGHTING_KEY "title_highlighting"
#define PREFERENCES_HIGHLIGHT_COLOR_KEY "title_highlighting_color"
#define PREFERENCES_BORDER_STYLE_KEY "title_border_style"
#define PREFERENCES_TEXT_RGBA_KEY "title_text_rgba"
#define PREFERENCES_SHADOW_RGBA_KEY "title_shadow_rgba"
#define PREFERENCES_BACKGROND_KEY "title_background"
#define PREFERENCES_GRADIENT_KEY "title_gradient"
#define PREFERENCES_SCALE_ENABLE_KEY "title_scale_enable"
#define PREFERENCES_BORDER_RGBA_KEY "title_border_rgba"
#define PREFERENCES_BORDER_WIDTH_KEY "title_border_width"
#define PREFERENCES_HIDE_DELAY_KEY "title_hide_delay"
#define PREFERENCES_FONT_KEY "title_font"

// preference defaults
#define PREFERENCES_ENABLE_DEFAULT "TRUE"
#define PREFERENCES_HIGHLIGHTING_DEFAULT "FALSE"
#define PREFERENCES_COLOR_DEFAULT "#000000AA"
#define PREFERENCES_BORDER_STYLE_DEFAULT "rounded"
#define PREFERENCES_TEXT_RGBA_DEFAULT "#FFFFFFFF"
#define PREFERENCES_SHADOW_RGBA_DEFAULT "#FFFFFF88"
#define PREFERENCES_BACKGROND_DEFAULT "gradient"
#define PREFERENCES_GRADIENT_DEFAULT "[#082D7688]"
#define PREFERENCES_SCALE_ENABLE_DEFAULT "FALSE"
#define PREFERENCES_BORDER_RGBA_DEFAULT "#145BC488"
#define PREFERENCES_BORDER_WIDTH_DEFAULT "1.0"
#define PREFERENCES_HIDE_DELAY_DEFAULT "800.0"
#define PREFERENCES_FONT_DEFAULT "Sans Regular 12"

/**
 * Title state.
 **/
typedef enum {
	TITLE_HIDING = 0,
	TITLE_HIDDEN,
	TITLE_SHOWING,
	TITLE_SHOWN
}KibaTitleState;

typedef struct {
	KibaTitle *title;
	GtkWidget *widget;

	CairoSurfaceData surface_data; /**< The title surface */
	PangoLayout *pango_layout;     /**< The pango layout of current text */
	char *text;                    /**< The current text */
	
	gint position;
}KibaTitleObject;

struct _KibaTitle {
	KibaTransparentWindow parent;

	GHashTable *objects;
	KibaTitleObject *active_object;

	gint flags; /**< Title flags */
	KibaTitleState state; /**< The title state */
	KibaTitleStyle style; /**< the background style */
	char *font; /**< The title font */
	gint hide_delay; /**< The delay to hide the title */

	float fade_alpha;
	gint scale_id;

	gint redraw_id;
	gboolean redraw_flag;

	GdkPixmap *pixmap; /**< offscreen pixmap */
	gint depth;
	gint diff;
	
	GtkOrientation orientation;
};

struct _KibaTitleClass {
	KibaTransparentWindowClass parent_class;
};

G_DEFINE_TYPE (KibaTitle, kiba_title, KIBA_TYPE_TRANSPARENT_WINDOW)

static PrefMember title_options[] = {
	{ PREFERENCES_ENABLE_KEY,
		PREFERENCES_ENABLE_DEFAULT, KIBA_BOOLEAN },
	{ PREFERENCES_HIGHLIGHTING_KEY,
		PREFERENCES_HIGHLIGHTING_DEFAULT, KIBA_BOOLEAN },
	{ PREFERENCES_HIGHLIGHT_COLOR_KEY,
		PREFERENCES_COLOR_DEFAULT, KIBA_STRING },
	{ PREFERENCES_BORDER_STYLE_KEY,
		PREFERENCES_BORDER_STYLE_DEFAULT, KIBA_STRING },
	{ PREFERENCES_TEXT_RGBA_KEY,
		PREFERENCES_TEXT_RGBA_DEFAULT, KIBA_STRING },
	{ PREFERENCES_SHADOW_RGBA_KEY,
		PREFERENCES_SHADOW_RGBA_DEFAULT, KIBA_STRING },
	{ PREFERENCES_BACKGROND_KEY,
		PREFERENCES_BACKGROND_DEFAULT, KIBA_STRING },
	{ PREFERENCES_GRADIENT_KEY,
		PREFERENCES_GRADIENT_DEFAULT, KIBA_LIST },
	{ PREFERENCES_SCALE_ENABLE_KEY,
		PREFERENCES_SCALE_ENABLE_DEFAULT, KIBA_BOOLEAN },
	{ PREFERENCES_BORDER_RGBA_KEY,
		PREFERENCES_BORDER_RGBA_DEFAULT, KIBA_STRING },
	{ PREFERENCES_BORDER_WIDTH_KEY,
		PREFERENCES_BORDER_WIDTH_DEFAULT, KIBA_DOUBLE },
	{ PREFERENCES_HIDE_DELAY_KEY,
		PREFERENCES_HIDE_DELAY_DEFAULT, KIBA_DOUBLE },
	{ PREFERENCES_FONT_KEY,
		PREFERENCES_FONT_DEFAULT, KIBA_STRING }
};

/// private methods

static void __resize(KibaTitle *title)
{
	GtkWidget *widget = GTK_WIDGET(title);
	GtkRequisition requisition = { 0, 0 };
	GtkAllocation allocation = { 0, 0, 0, 0 };
	
	gtk_widget_size_request (widget, &requisition);
	allocation.width = requisition.width;
	allocation.height = requisition.height;
	gtk_widget_size_allocate (widget, &allocation);
}

static cairo_t* __cairo_create (KibaTitle *title)
{
	GtkWidget *widget;
	cairo_t *cr;

	if (!GTK_WIDGET_REALIZED(title))
		return NULL;

	widget = GTK_WIDGET(title);
	
	cr = gdk_cairo_create (title->pixmap);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS)
		return cr; // drawing offscreen
	// fallback to window surface
	return gdk_cairo_create (widget->window);
}

static gint __redraw (gpointer data)
{
	KibaTitle *title = data;
	KibaTitleObject *object = title->active_object;
	GtkWidget *widget = GTK_WIDGET(title);
	cairo_t *cr;
	int x, y;
	GdkGC *gc;
	gint i;

	if (!object || !PANGO_IS_LAYOUT(object->pango_layout))
		return;
	
	const KibaLayoutData *d = &kiba_layout_data[object->position];
	
	// reload offscreen pixmap
	if (GDK_IS_PIXMAP(title->pixmap))
		g_object_unref(title->pixmap);
	title->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width,
		widget->allocation.height, title->depth);
	
	// create cairo context
	cr = __cairo_create(title);
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		return;
	
	// clear background to be transparent
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	if (kiba_transparent_window_is_composited(
	    KIBA_TRANSPARENT_WINDOW(title))) {
		kiba_transparent_window_draw_background(
			KIBA_TRANSPARENT_WINDOW(title), cr);
	} else {
		cairo_set_source_rgba (cr, 0, 0, 0, 0);
		cairo_paint (cr);
	}

	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	
	x = (widget->allocation.width-object->surface_data.width)/2;
	y = (widget->allocation.height-object->surface_data.height)/2;
	// make sure the surface is on the window
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		y = MAX(0, MIN(widget->allocation.height-object->surface_data.height, y-title->diff));
	} else {
		x = MAX(0, MIN(widget->allocation.width-object->surface_data.width, x-title->diff));
	}
	
	cairo_translate(cr, x, y);

	if (GET_FLAG(title->flags, KIBA_TITLE_SCALING)) {
		float fade_factor = title->fade_alpha < 0.8f ? title->fade_alpha + 0.2f : 1.0f;
		cairo_translate (cr,
			-((fade_factor-1.0f) * object->surface_data.width) / 2.0f,
			-((fade_factor-1.0f) * object->surface_data.height) / 2.0f);
		cairo_scale (cr, fade_factor, fade_factor);
	}
	
	// rotate the title to be vertical
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		cairo_translate (cr, (d->halignment-1)*
			(widget->allocation.width-object->surface_data.width)/2, 0);
	}

	cairo_save (cr);

	cairo_translate (cr,
		title->style.border_width + 5,
		title->style.border_width + 1);

	// layout the shadow
	cairo_set_source_rgba (cr,
			title->fade_alpha*title->style.shadow_rgba.r,
			title->fade_alpha*title->style.shadow_rgba.g,
			title->fade_alpha*title->style.shadow_rgba.b,
			title->fade_alpha*title->style.shadow_rgba.a);
	for (i = 0; i < 4; i++) {
		cairo_save (cr);
			cairo_translate (cr, (1 - (i&2)), (1 - 2*(i&1)));
			pango_cairo_show_layout (cr, object->pango_layout);
		cairo_restore (cr);
	}

	// layout the text
	cairo_set_source_rgba (cr, title->fade_alpha*title->style.rgba.r,
		title->fade_alpha*title->style.rgba.g,
		title->fade_alpha*title->style.rgba.b,
		title->fade_alpha*title->style.rgba.a);
	cairo_save (cr);
		pango_cairo_show_layout (cr, object->pango_layout);
	cairo_restore (cr);
	cairo_restore (cr);

	// draw highlight
	// FIXME: heavy cpu load 
	if (GET_FLAG(title->flags, KIBA_TITLE_HIGHLIGHTING)) {
		cairo_save (cr);
			cairo_pattern_t *pat;
			cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);

			static float title_highlight_offset = -1.0f;
			title_highlight_offset += 0.01f;
			if (title_highlight_offset > 1.0f)
				title_highlight_offset = -1.0f;

			if (d->orientation == GTK_ORIENTATION_VERTICAL) {
				pat = cairo_pattern_create_linear (
					0.0,
					object->surface_data.height * title_highlight_offset,
					object->surface_data.width,
					object->surface_data.height * (title_highlight_offset + 0.7));
			} else {
				pat = cairo_pattern_create_linear (
					object->surface_data.width * title_highlight_offset,
					0.0,
					object->surface_data.width * (title_highlight_offset + 0.7),
					object->surface_data.height);
			}
			cairo_pattern_add_color_stop_rgba (pat, 0.9, 1.0, 1.0, 1.0, 0.0);
			cairo_pattern_add_color_stop_rgba (pat, 0.3,
				title->fade_alpha*title->style.highlightColor.r,
				title->fade_alpha*title->style.highlightColor.g,
				title->fade_alpha*title->style.highlightColor.b,
				title->fade_alpha*title->style.highlightColor.a);
			cairo_pattern_add_color_stop_rgba (pat, 0.0, 1.0, 1.0, 1.0, 0.0);
			cairo_set_source (cr, pat);
			cairo_paint(cr);
			cairo_pattern_destroy (pat);

			/* queue draw as long as the widget is visible */
			if (title->fade_alpha > 0.0f) {
				title->redraw_flag = TRUE;
				gtk_widget_queue_draw(GTK_WIDGET(title));
			}
		cairo_restore (cr);
	}

	// draw background
	cairo_save (cr);
		cairo_set_operator (cr, CAIRO_OPERATOR_DEST_OVER);
		cairo_set_source_surface (cr, object->surface_data.surface, 0.0, 0.0);
		cairo_paint_with_alpha (cr, title->fade_alpha);
	cairo_restore (cr);
	
	// clear background to be transparent
	if (!kiba_transparent_window_is_composited(KIBA_TRANSPARENT_WINDOW(title))) {
		cairo_save (cr);
			cairo_set_operator (cr, CAIRO_OPERATOR_DEST_OVER);
			kiba_transparent_window_draw_background(
				KIBA_TRANSPARENT_WINDOW(title), cr);
		cairo_restore (cr);
	}

	cairo_destroy(cr);
	
	// render to the window
	gc = gdk_gc_new (title->pixmap);
	gdk_draw_drawable (widget->window,
			   gc,
			   title->pixmap,
			   0, 0, 0, 0,
			   widget->allocation.width,
			   widget->allocation.height);
	gdk_gc_unref (gc);

	title->redraw_id = 0;

	return FALSE;
}

static void __draw_object_surface (cairo_t *cr,
		cairo_surface_t  *surface,
		gint width,
		gint height,
		void *data)
{
	KibaTitleObject *object = data;
	cairo_pattern_t *pattern = NULL;

	cairo_paint_shape (cr,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_width/2,
		object->title->style.border_width/2,
		object->surface_data.width - object->title->style.border_width/2,
		object->surface_data.height - object->title->style.border_width/2,
		5);

	if (!object->title->style.file)
		pattern = cairo_get_gradient_pattern (object->title->style.border_width/2,
				object->title->style.border_width/2,
				object->surface_data.width - object->title->style.border_width/2,
				object->surface_data.height - object->title->style.border_width/2,
				object->title->style.gradient, object->title->style.numColors);
	else
		pattern = cairo_get_png_pattern (object->title->style.file, 1.0);
	if (pattern == NULL)
		return;

	cairo_set_source (cr, pattern);
	cairo_fill_preserve (cr);
	cairo_pattern_destroy (pattern);

	cairo_set_source_rgba (cr,
		object->title->style.border_rgba.r,
		object->title->style.border_rgba.g,
		object->title->style.border_rgba.b,
		object->title->style.border_rgba.a);
	cairo_set_line_width (cr, object->title->style.border_width);
	cairo_stroke (cr);
}

static void __remove_objects (gpointer key,
                              gpointer value,
                              gpointer user_data) {
	kiba_title_remove_object(KIBA_TITLE(user_data), key);
}

static void __update_object_pango_layout (gpointer key,
                                          gpointer value,
                                          gpointer user_data) {
	KibaTitle *title = user_data;
	KibaTitleObject *object = value;
	cairo_t *cr = __cairo_create (title);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS) {
		cairo_update_pango_layout (object->text,
			title->font,
			cr,
			&object->pango_layout);
		cairo_destroy (cr);
	}
}

static void __update_object_surface (gpointer key,
                                     gpointer value,
                                     gpointer user_data) {
	KibaTitle *title = user_data;
	KibaTitleObject *object = value;
	const KibaLayoutData *d;
	PangoRectangle  log;
	cairo_t *cr;

	d = &kiba_layout_data[object->position];

	if (!PANGO_IS_LAYOUT(object->pango_layout))
		return;
	
	struct {
		gint  *w,  *h;
	} magic[2] = {
		{ &object->surface_data.width,  &object->surface_data.height },
		{ &object->surface_data.height, &object->surface_data.width }
	};
	pango_layout_get_pixel_extents (object->pango_layout, NULL, &log);
	*magic[d->orientation].w = log.width + 10 + 2*object->title->style.border_width;
	*magic[d->orientation].h = log.height + 2 + 2*object->title->style.border_width;
	
	cr = __cairo_create (object->title);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS) {
		cairo_update_surface (cr, &object->surface_data, __draw_object_surface, object);
		cairo_destroy (cr);
	}
}

static gint __fade_timeout (gpointer data) {
	KibaTitle *title = data;
	
	title->redraw_flag = TRUE;
	gtk_widget_queue_draw (GTK_WIDGET(title));

	if (title->state == TITLE_HIDING) {
		title->fade_alpha -= 1.0f / title->hide_delay;

		if (title->fade_alpha <= 0.0f) {
			title->state = TITLE_HIDDEN;
			title->scale_id = 0;
			gtk_widget_hide(GTK_WIDGET(title));
			return FALSE;
		}
	}
	else {
		title->scale_id = 0;
		return FALSE;
	}

	return TRUE;
}

static void __object_size_allocate_cb (GtkWidget *widget,
		GtkAllocation *allocation, gpointer user_data)
{
	KibaTitleObject *object = user_data;

	if (!kiba_title_object_is_shown(object->title, object->widget))
		return;

	object->title->redraw_flag = TRUE;
	gtk_widget_queue_resize (GTK_WIDGET(object->title));
}

static void __object_destroy_cb (GtkWidget *widget,
                                 gpointer data) {
	kiba_title_remove_object(KIBA_TITLE(data), widget);
}

static void __value_destroy_func (gpointer value) {
	KibaTitleObject *title_object = value;
	
	if (title_object->title->active_object == title_object) {
		title_object->title->active_object = NULL;
	}

	if (GTK_IS_WIDGET(title_object->widget)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(title_object->widget),
			  G_CALLBACK(__object_size_allocate_cb), title_object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(title_object->widget),
			  G_CALLBACK(__object_destroy_cb), title_object->title);
	}

	if (title_object->surface_data.surface)
		cairo_surface_destroy (title_object->surface_data.surface);
	if (title_object->pango_layout)
		g_object_unref (title_object->pango_layout);
	if (title_object->text)
		g_free (title_object->text);

	g_free (title_object);
}

static void __get_largest (gpointer key,
                           gpointer value,
                           gpointer user_data) {
	GtkRequisition **requisition = user_data;
	KibaTitleObject *current = value;
	
	if (PANGO_IS_LAYOUT(current->pango_layout)) {
		PangoRectangle log;
		pango_layout_get_pixel_extents (current->pango_layout, NULL, &log);
		(*requisition)->width  = MAX((*requisition)->width, log.width + 10);
		(*requisition)->height = MAX((*requisition)->height, log.height + 2);
	}
}

static void
prefs_callback(const gchar *option_id,
		gpointer value,
		gpointer data)
{
	KibaTitle *title = data;
	
	if (strcmp(option_id, PREFERENCES_ENABLE_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_title_set_flag (title, KIBA_TITLE_DRAWABLE);
		} else {
			kiba_title_unset_flag (title, KIBA_TITLE_DRAWABLE);
		}
	} else if (strcmp(option_id, PREFERENCES_HIGHLIGHTING_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_title_set_flag (title, KIBA_TITLE_HIGHLIGHTING);
		} else {
			kiba_title_unset_flag (title, KIBA_TITLE_HIGHLIGHTING);
		}
	} else if (strcmp(option_id, PREFERENCES_SCALE_ENABLE_KEY) == 0) {
		if ((gboolean)(GPOINTER_TO_INT (value)) == TRUE) {
			kiba_title_set_flag (title, KIBA_TITLE_SCALING);
		} else {
			kiba_title_unset_flag (title, KIBA_TITLE_SCALING);
		}
	} else if (strcmp(option_id, PREFERENCES_HIGHLIGHT_COLOR_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		style->highlightColor = kiba_prefs_rgba_from_string ((gchar*) value);
	} else if (strcmp(option_id, PREFERENCES_TEXT_RGBA_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		style->rgba = kiba_prefs_rgba_from_string ((gchar*) value);
	} else if (strcmp(option_id, PREFERENCES_SHADOW_RGBA_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		style->shadow_rgba =
			kiba_prefs_rgba_from_string ((gchar*) value);
	} else if (strcmp(option_id, PREFERENCES_BACKGROND_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		if (g_str_has_suffix ((gchar*) value, ".png") ||
		    g_str_has_suffix ((gchar*) value, ".svg")) {
			style->file = (gchar*) value;
		} else {
			style->file = NULL;
		}
		kiba_title_set_style (title, *style);
	} else if (strcmp(option_id, PREFERENCES_GRADIENT_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		GSList *list = value;
		gint length = g_slist_length(list);
		KibaRGBA *gradient = g_new0(KibaRGBA, length);
		
		kiba_prefs_string_list_to_rgba_array(list, gradient);
		if (style->gradient) {
			g_free(style->gradient);
		}
		style->gradient = gradient;
		style->numColors = length;
		kiba_title_set_style (title, *style);
	} else if (strcmp(option_id, PREFERENCES_BORDER_STYLE_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		style->border_style = kiba_prefs_edge_style_from_string ((gchar*) value);
		kiba_title_set_style (title, *style);
	} else if (strcmp(option_id, PREFERENCES_BORDER_RGBA_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		style->border_rgba = kiba_prefs_rgba_from_string ((gchar*) value);
		kiba_title_set_style (title, *style);
	} else if (strcmp(option_id, PREFERENCES_BORDER_WIDTH_KEY) == 0) {
		KibaTitleStyle *style = kiba_title_get_style(title);
		style->border_width = (gfloat)g_ascii_strtod((gchar*) value, NULL);
		kiba_title_set_style (title, *style);
	} else if (strcmp(option_id, PREFERENCES_HIDE_DELAY_KEY) == 0) {
		kiba_title_set_hide_delay(title,
			(gint) (gfloat)g_ascii_strtod((gchar*) value, NULL));
	} else if (strcmp(option_id, PREFERENCES_FONT_KEY) == 0) {
		kiba_title_set_font (title, (gchar*) value);
	}
}

/// Public methods

/**
 * kiba_title_object_add:
 * @title: #KibaTitle instance.
 * @widget: #GtkWidget associated to @title
 * @position: position of @widget.
 *
 * Adds title object.
 */
void kiba_title_add_object (KibaTitle *title,
		GtkWidget *widget, KibaPosition position)
{
	KibaTitleObject *title_object;

	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (GTK_IS_WIDGET(widget));

	title_object = g_hash_table_lookup (title->objects, widget);
	g_return_if_fail(title_object == NULL);

	title_object = g_new0 (KibaTitleObject, 1);
	title_object->position = position;
	title_object->widget = widget;
	title_object->pango_layout = NULL;
	title_object->text = NULL;
	title_object->surface_data.surface = NULL;
	title_object->surface_data.width = 1;
	title_object->surface_data.height = 1;
	title_object->title = title;

	g_hash_table_insert (title->objects, widget, title_object);

	g_signal_connect_after (G_OBJECT(widget), "size-allocate",
		G_CALLBACK(__object_size_allocate_cb), title_object);
	g_signal_connect (G_OBJECT(widget), "destroy",
		G_CALLBACK(__object_destroy_cb), title);
}

/**
 * kiba_title_remove_object:
 * @title: #KibaTitle instance.
 * @widget: #GtkWidget associated to @title
 *
 * Removes title object.
 */
void kiba_title_remove_object (KibaTitle *title, GtkWidget *widget)
{
	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (GTK_IS_WIDGET(widget));
	
	g_hash_table_remove(title->objects, widget);
}

/**
 * kiba_title_show_object:
 * @title: #KibaTitle instance.
 * @widget: #GtkWidget associated to @title
 * @position: position of @widget.
 *
 * Shows title object.
 */
void kiba_title_show_object (KibaTitle *title,
		GtkWidget *widget, KibaPosition position)
{
	const KibaLayoutData *d;
	gboolean show_widget;
	KibaTitleObject *title_object;

	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (GTK_IS_WIDGET(widget));

	title_object = g_hash_table_lookup (title->objects, widget);
	g_return_if_fail (title_object != NULL);

	if (title_object->pango_layout == NULL)
		return;

	d = &kiba_layout_data[position];

	if (!title->active_object) {
		show_widget = TRUE;
	}
	
	title_object->position = position;
	title->active_object = title_object;
	title_object->title->state = TITLE_SHOWING;

	if (title->scale_id) {
		g_source_remove(title->scale_id);
		title->scale_id = 0;
	}
	title->fade_alpha = 1.0f;
	title_object->title->state = TITLE_SHOWN;

	title->redraw_flag = TRUE;
	if (show_widget)
		gtk_widget_show(GTK_WIDGET(title));
	gtk_widget_queue_resize(GTK_WIDGET(title));
	gdk_window_raise (GTK_WIDGET(title)->window);
}

/**
 * kiba_title_hide_object:
 * @title: #KibaTitle instance.
 * @widget: #GtkWidget associated to @title
 *
 * Hides title object.
 */
void kiba_title_hide_object (KibaTitle *title, GtkWidget *widget)
{
	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (GTK_IS_WIDGET(widget));
	
	KibaTitleObject *title_object;

	title_object = g_hash_table_lookup (title->objects, widget);
	g_return_if_fail (title_object != NULL);

	if (title_object == title->active_object) {
	    if (!title->scale_id)
		    title->scale_id = g_timeout_add (1000/FADE_FPS,
				__fade_timeout, title_object->title);
	    title_object->title->state = TITLE_HIDING;
	}
}

/**
 * kiba_title_object_is_shown:
 * @title: #KibaTitle instance.
 * @widget: #GtkWidget associated to @title
 *
 * Returns: wether the title is shown.
 */
gboolean kiba_title_object_is_shown (KibaTitle *title, GtkWidget *widget)
{
	g_return_val_if_fail (KIBA_IS_TITLE(title), FALSE);
	g_return_val_if_fail (GTK_IS_WIDGET(widget), FALSE);
	
	KibaTitleObject *title_object;

	title_object = g_hash_table_lookup (title->objects, widget);
	g_return_val_if_fail (title_object != NULL, FALSE);

	return (title_object == title->active_object);
}

/**
 * kiba_title_set_object_text:
 * @title: KibaTitle instance
 * @widget: The widget to set the text for
 * @text: the text to set
 *
 * Sets title text.
 */
void kiba_title_set_object_text (KibaTitle *title,
		GtkWidget *widget,
		const char *text)
{
	KibaTitleObject *title_object;
	cairo_t *cr;
	gint diff;

	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (GTK_IS_WIDGET(widget));

	title_object = g_hash_table_lookup (title->objects, widget);

	g_return_if_fail (title_object != NULL);
	g_return_if_fail (text != NULL);

	if (title_object->text) {
		if (strcmp (title_object->text, text) == 0)
			return;
		diff = strlen (title_object->text) - strlen (text);
		g_free (title_object->text);
	} else {
		diff = -5;
	}
	title_object->text = g_strdup (text);
	
	// update layout and surface
	__update_object_pango_layout(widget, title_object, title);
	if (diff != 0)
		__update_object_surface (widget, title_object, title);

	gtk_widget_queue_resize_no_redraw (GTK_WIDGET(title));
}

/**
 * kiba_title_get_object_text:
 * @title: #KibaTitle instance.
 * @widget: #GtkWidget associated to @title
 *
 * Returns: title text.
 */
const char* kiba_title_get_object_text (KibaTitle *title, GtkWidget *widget)
{
	g_return_val_if_fail (KIBA_IS_TITLE(title), NULL);
	g_return_val_if_fail (GTK_IS_WIDGET(widget), NULL);
	
	KibaTitleObject *title_object;

	title_object = g_hash_table_lookup (title->objects, widget);
	g_return_val_if_fail (title_object != NULL, NULL);

	return title_object->text;
}

/**
 * kiba_title_set_font:
 * @title: #KibaTitle instance.
 * @font: font for @title.
 *
 * Sets title font.
 */
void kiba_title_set_font (KibaTitle *title, char *font)
{
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	GtkWidget *widget = GTK_WIDGET(title);
	GList *l;

	title->font = font;

	/* update pango layouts for the new font */
	if (title->objects) {
		g_hash_table_foreach (title->objects,
			__update_object_pango_layout, title);
		g_hash_table_foreach (title->objects,
			__update_object_surface, title);
	}
	
	title->redraw_flag = TRUE;
	gtk_widget_queue_resize(GTK_WIDGET(title));
	__resize (title);
}

/**
 * kiba_title_set_flag:
 * @title: #KibaTitle instance.
 * @flag: flag to set.
 *
 * Sets title flag.
 */
void kiba_title_set_flag (KibaTitle *title, KibaTitleFlag flag)
{
	g_return_if_fail (KIBA_IS_TITLE(title));

	if (GET_FLAG(title->flags, flag))
		return;

	SET_FLAG(title->flags, flag);

	switch (flag) {
	default:
		break;
	}
}

/**
 * kiba_title_unset_flag:
 * @title: #KibaTitle instance.
 * @flag: flag to unset.
 *
 * Unsets title flag.
 */
void kiba_title_unset_flag (KibaTitle *title,  KibaTitleFlag flag) 
{
	g_return_if_fail (KIBA_IS_TITLE(title));

	if (!GET_FLAG(title->flags, flag))
		return;

	DEL_FLAG(title->flags, flag);

	switch (flag) {
	default:
		break;
	}
}

/**
 * kiba_title_get_flag:
 * @title: #KibaTitle instance.
 * @flag: flag to lookup.
 *
 * Returns: TRUE if @flag is set, otherwhise FALSE is returned.
 */
gboolean kiba_title_get_flag (KibaTitle *title, KibaTitleFlag flag)
{
	g_return_val_if_fail(KIBA_IS_TITLE(title), FALSE);
	return GET_FLAG(title->flags, flag);
}

/**
 * kiba_title_set_style:
 * @title: #KibaTitle instance.
 * @style: @title style.
 *
 * Sets the background style.
 */
void kiba_title_set_style (KibaTitle *title, KibaTitleStyle style)
{
	GList *l;

	g_return_if_fail (KIBA_IS_TITLE(title));

	title->style = style;
	
	if (title->objects) {
		g_hash_table_foreach (title->objects,
			__update_object_surface, title);
	}

	if (kiba_title_is_shown (title)) {
		gtk_widget_queue_draw(GTK_WIDGET(title));
		title->redraw_flag = TRUE;
	}
}

/**
 * kiba_title_is_shown:
 * @title: #KibaTitle instance.
 *
 * Returns: wether the title is shown.
 */
gboolean kiba_title_is_shown (KibaTitle *title)
{
	g_return_val_if_fail (KIBA_IS_TITLE(title), FALSE);
	return title->active_object != NULL;
}

/**
 * kiba_title_get_style:
 * @title: #KibaTitle instance.
 *
 * Returns: @title style.
 */
KibaTitleStyle* kiba_title_get_style (KibaTitle *title)
{
	g_return_val_if_fail(KIBA_IS_TITLE(title), NULL);
	return &title->style;
}

/**
 * kiba_title_set_hide_delay:
 * @title: #KibaTitle instance.
 * @hide_delay: amount of milliseconds to wait before hiding the title.
 */
void kiba_title_set_hide_delay (KibaTitle *title, gint hide_delay)
{
	g_return_if_fail(KIBA_IS_TITLE(title));
	title->hide_delay = hide_delay;
}

/// Widget Class Implementations

/*
 * gtk_widget_expose implementation.
 */
static gboolean kiba_title_expose_event (GtkWidget *widget,
                                         GdkEventExpose *event) {
	KibaTitle *title = KIBA_TITLE(widget);

	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;

	if (!title->redraw_id && title->redraw_flag) {
		title->redraw_flag = FALSE;
		title->redraw_id = g_timeout_add (1000/MAX_FPS, __redraw, widget);
	}

	return FALSE;
}

/*
 * gtk_widget_screen_changed implementation.
 */
static void kiba_title_screen_changed (GtkWidget *widget,
                                       GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;

	screen = gtk_widget_get_screen (widget);

	colormap = gdk_screen_get_rgba_colormap (screen);
	visual = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual = gdk_screen_get_rgb_visual (screen);
	}

	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

/*
 * gtk_widget_size_allocate implementation.
 */
static void kiba_title_size_allocate (GtkWidget *widget,
                                      GtkAllocation *allocation)  {
	KibaTitle *title = KIBA_TITLE(widget);
	KibaTitleObject *object = title->active_object;
	gboolean size_changed = FALSE;
	GtkWidget *parent;
	gint buf;

	if (allocation->width != widget->allocation.width ||
	    allocation->height != widget->allocation.height)
		size_changed = TRUE;
	
	if (!object) {
		widget->allocation = *allocation;
		return;
	}
	
	const KibaLayoutData *d = &kiba_layout_data[object->position];
	
	// get title position
	allocation->x = object->widget->allocation.x;
	allocation->y = object->widget->allocation.y;
	for (parent = object->widget; parent->parent != NULL; parent = parent->parent);
	if (parent != object->widget) {
		allocation->x += parent->allocation.x;
		allocation->y += parent->allocation.y;
	}
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation->x += (d->halignment == 0 ? object->widget->allocation.width :
			-allocation->width) - (d->halignment - 1) * TITLE_OBJECT_DISTANCE;
		allocation->y += -(allocation->height - object->widget->allocation.height)/2;
		buf = allocation->y;
	}
	else {
		allocation->x += -(allocation->width - object->widget->allocation.width)/2;
		allocation->y += (d->valignment == 0 ? object->widget->allocation.height :
			-allocation->height) - (d->valignment - 1) * TITLE_OBJECT_DISTANCE;
		buf = allocation->x;
	}
	allocation->x = MIN(
		KIBA_SCREEN_DEFAULT->geometry.x +
		KIBA_SCREEN_DEFAULT->geometry.width - allocation->width,
		MAX(KIBA_SCREEN_DEFAULT->geometry.x, allocation->x));
	allocation->y = MIN(
		KIBA_SCREEN_DEFAULT->geometry.y +
		KIBA_SCREEN_DEFAULT->geometry.height - allocation->height,
		MAX(KIBA_SCREEN_DEFAULT->geometry.y, allocation->y));
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		title->diff = allocation->y-buf;
	} else {
		title->diff = allocation->x-buf;
	}
	
	widget->allocation = *allocation;
	
	if (size_changed) {
		if (GDK_IS_PIXMAP(title->pixmap))
			g_object_unref(title->pixmap);
		title->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width,
				widget->allocation.height, title->depth);
		
		gdk_window_move_resize(widget->window,
			allocation->x, allocation->y,
			allocation->width, allocation->height);
		// redraw immediate
		__redraw(widget);
	} else {
		gdk_window_move(widget->window,
			allocation->x, allocation->y);
	}
}

/*
 * gtk_widget_size_request implementation.
 */
static void kiba_title_size_request(GtkWidget *widget,
		GtkRequisition *requisition)
{
	KibaTitle *title = KIBA_TITLE(widget);
	
	requisition->width = 0;
	requisition->height = 0;
	if (title->objects) {
		g_hash_table_foreach (title->objects,
			__get_largest,
			&requisition);
	}
	requisition->width += 2*title->style.border_width;
	requisition->height += 2*title->style.border_width;
}

/*
 * gtk_widget_unrealize implementation.
 */
static void kiba_title_unrealize (GtkWidget *widget) {
	KibaTitle *title = KIBA_TITLE(widget);

	if (GDK_IS_PIXMAP(title->pixmap))
		g_object_unref(title->pixmap);
	title->pixmap = NULL;

	g_hash_table_remove_all (title->objects);
	g_hash_table_unref(title->objects);
	title->objects = NULL;

	(* GTK_WIDGET_CLASS (kiba_title_parent_class)->unrealize) (widget);
}

/*
 * gtk_widget_realize implementation.
 */
static void kiba_title_realize (GtkWidget *widget) {
	KibaTitle *title = KIBA_TITLE(widget);
	GtkWindow *window = GTK_WINDOW (widget);
	GdkWindowAttr attributes;
	gint attributes_mask;

	/* make sure the size is properly allocated */
	if (widget->allocation.x == -1 &&
	    widget->allocation.y == -1 &&
	    widget->allocation.width == 1 &&
	    widget->allocation.height == 1) {
		GtkRequisition requisition;
		GtkAllocation allocation = { 0, 0, 0, 0 };

		gtk_widget_size_request (widget, &requisition);
		allocation.width = requisition.width;
		allocation.height = requisition.height;
		gtk_widget_size_allocate (widget, &allocation);
	}

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

	attributes.window_type = GDK_WINDOW_TOPLEVEL;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.title = window->title;
	attributes.wmclass_name = window->wmclass_name;
	attributes.wmclass_class = window->wmclass_class;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.visual = gtk_widget_get_visual(widget);
	attributes.colormap = gtk_widget_get_colormap(widget);
	attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	if (attributes.visual != NULL && attributes.colormap != NULL)
		attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;
	attributes_mask |= (window->title ? GDK_WA_TITLE : 0);
	attributes_mask |= (window->wmclass_name ? GDK_WA_WMCLASS : 0);

	widget->window = gdk_window_new (gdk_get_default_root_window (),
				&attributes, attributes_mask);
	/* catch the window events on the widget */
	gdk_window_set_user_data (widget->window, window);
	widget->style = gtk_style_attach (widget->style, widget->window);

	gdk_window_set_type_hint (widget->window, GDK_WINDOW_TYPE_HINT_DOCK);
	gdk_window_set_icon_name (widget->window, WINDOW_ICON_NAME);
	gdk_window_stick (widget->window);
	gdk_window_set_keep_above (widget->window, TRUE);

	title->depth = attributes.visual->depth;
	title->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width,
			widget->allocation.height, title->depth);

	title->objects = g_hash_table_new_full (
		NULL, /**<< using default hash function **/
		g_direct_equal, /**<< comparing pointers **/
		NULL,
		__value_destroy_func);
}

/// Class init

static void kiba_title_class_init (KibaTitleClass *class) {
	GtkWidgetClass *widget_class;

	widget_class = GTK_WIDGET_CLASS (class);

	widget_class->realize = kiba_title_realize;
	widget_class->unrealize = kiba_title_unrealize;
	widget_class->size_allocate = kiba_title_size_allocate;
	widget_class->size_request = kiba_title_size_request;
	widget_class->expose_event = kiba_title_expose_event;
	widget_class->screen_changed = kiba_title_screen_changed;
}

static void kiba_title_init (KibaTitle *title) {
	KibaScreen *screen;
	KibaPrefsFile* general_file;
	gint i;

	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (title),
			GTK_NO_WINDOW |
			GTK_DOUBLE_BUFFERED);

	screen = kiba_screen_get();

	gtk_widget_add_events (GTK_WIDGET (title),
		GDK_EXPOSURE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	gtk_widget_set_app_paintable (GTK_WIDGET (title), TRUE);
	gtk_window_set_role (GTK_WINDOW(title), WINDOW_ROLE);
	
	SET_FLAG(title->flags, KIBA_TITLE_DRAWABLE);

	KibaRGBA *gradient = g_new0(KibaRGBA, 1);
	gradient[0] = (KibaRGBA) {1.0,1.0,1.0,1.0};
	title->style = (KibaTitleStyle) {
		NULL, KIBA_ROUNDED,
		2.0f, 45.0f, 5.0f,
		{ 1.0f, 1.0f, 1.0f, 0.8f },
		{ 0.0f, 0.0f, 0.0f, 0.8f },
		{ 0.8f, 0.8f, 0.8f, 0.8f }, { 0.0f, 0.0f, 0.0f, 0.8f },
		gradient, 1
	};
	title->font = "Arial bold 12";
	title->hide_delay = 50;
	title->state = TITLE_HIDDEN;
	title->objects = NULL;
	title->active_object = NULL;
	title->fade_alpha = 0.0f;
	title->scale_id = 0;
	title->redraw_id = 0;
	title->redraw_flag = TRUE;
	
	// add preferences
	general_file = kiba_prefs_get_general_file (kiba_prefs_get());
	kiba_prefs_add_options (general_file,
		title_options, G_N_ELEMENTS(title_options));
	for (i = 0; i < G_N_ELEMENTS(title_options); i++) {
		kiba_prefs_add_callback (general_file,
			title_options[i].id, prefs_callback, title, TRUE);
	}

	kiba_title_screen_changed(GTK_WIDGET(title), screen->gdk_screen);
}

/**
 * kiba_title_new:
 *
 * Creates and returns a new KibaTitle.
 *
 * Returns: The KibaTitle object created.
 */
KibaTitle* kiba_title_new (void) {
	return g_object_new (KIBA_TYPE_TITLE, NULL);
}
